from typing import Sequence, Optional, Any, List

import numpy as np
from matplotlib import pyplot as plt
from scipy.special import softmax


def plot_action_values(stats: Sequence[Sequence[float]],
                       action_tags: Optional[Sequence[Any]] = None,
                       size: float = 5.,
                       titles: List[str] = None):
    # Plot stackplot of how action expectations and probabilities changed
    x = np.arange(len(stats))
    y = np.array(stats).T
    y_prob = softmax(y, axis=0)

    labels = [str(action) for action in action_tags]
    fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(size * 2, size))
    ax0, ax1 = ax

    for ys, label in zip(y, labels):
        ax0.plot(x, ys, label=label)
        ax0.grid()
    ax1.stackplot(x, y_prob, labels=labels)

    if not titles:
        expectation_values_title = 'Action Expectation Values'
        probabilities_title = 'Action Probabilities'
    else:
        expectation_values_title, probabilities_title = titles
    ax0.set_title(expectation_values_title, size=10)
    ax0.set_xlabel('Generation')
    ax0.set_ylabel('Reward Expectation')
    ax1.set_title(probabilities_title, size=10)
    ax1.set_xlabel('Generation')
    ax1.set_ylabel('Probability')
    ax1.set(ylim=(0, 1.0), yticks=np.linspace(0., 1., 21))

    if action_tags:
        ax[0].legend(loc='upper right')



import os
import platform
import tempfile

from pathlib import Path
from typing import Callable


def copy_doc(source_func: Callable) -> Callable:
    """
    Copies a docstring from the provided ``source_func`` to the wrapped function

    :param source_func: function to copy the docstring from

    :return: wrapped function with the same docstring as in the given ``source_func``
    """
    def wrapper(func: Callable) -> Callable:
        func.__doc__ = source_func.__doc__
        return func
    return wrapper


def project_root() -> Path:
    """Returns project root folder."""
    return Path(__file__).parent.parent.parent


def default_data_dir() -> str:
    """ Returns the folder where all the output data is recorded to
    (logs, history, visualisations). Default is: `/tmp/GOLEM`
    """
    name = 'GOLEM'
    temp_folder = Path("/tmp" if platform.system() == "Darwin" else tempfile.gettempdir())

    default_data_path = os.path.join(temp_folder, name)

    if name not in os.listdir(temp_folder):
        os.mkdir(default_data_path)

    return default_data_path



from typing import Any, Dict, Type, Sequence, Union

from golem.core.dag.graph import Graph
from golem.core.dag.graph_delegate import GraphDelegate
from golem.core.dag.linked_graph_node import LinkedGraphNode


def graph_from_json(cls: Type[Graph], json_obj: Dict[str, Any]) -> Graph:
    obj = cls()

    nodes_key = 'nodes' if 'nodes' in json_obj else '_nodes'
    if not issubclass(cls, GraphDelegate):
        nodes = json_obj[nodes_key]
        _reassign_edges_by_node_ids(nodes)
        obj.nodes = nodes

    # GraphDelegate case is handled by this
    vars(obj).update(**{k: v for k, v in json_obj.items() if k != nodes_key})

    return obj


def _reassign_edges_by_node_ids(nodes: Sequence[Union[LinkedGraphNode, dict]]):
    """
    Assigns each <inner_node> from "nodes_from" to equal <outer_node> from "nodes"
        (cause each node from "nodes_from" in fact should point to the same node from "nodes")
    """
    lookup_dict = {}
    for node in nodes:
        if isinstance(node, dict):
            lookup_dict[node['uid']] = node
        else:
            lookup_dict[node.uid] = node

    for node in nodes:
        nodes_from = node['_nodes_from'] if isinstance(node, dict) else node.nodes_from
        if not nodes_from:
            continue
        for parent_node_idx, parent_node_uid in enumerate(nodes_from):
            nodes_from[parent_node_idx] = lookup_dict.get(parent_node_uid, None)



import os
from typing import Optional, Union

import seaborn as sns
from matplotlib import pyplot as plt, ticker

from golem.visualisation.opt_history.utils import get_history_dataframe, show_or_save_figure
from golem.visualisation.opt_history.history_visualization import HistoryVisualization


class FitnessBox(HistoryVisualization):
    def visualize(self, save_path: Optional[Union[os.PathLike, str]] = None,
                  dpi: Optional[int] = None, best_fraction: Optional[float] = None):
        """ Visualizes fitness values across generations in the form of boxplot.

        :param save_path: path to save the visualization. If set, then the image will be saved, and if not,
            it will be displayed.
        :param dpi: DPI of the output figure.
        :param best_fraction: fraction of the best individuals of each generation that included in the
            visualization. Must be in the interval (0, 1].
        """
        save_path = save_path or self.get_predefined_value('save_path')
        dpi = dpi or self.get_predefined_value('dpi')
        best_fraction = best_fraction or self.get_predefined_value('best_fraction')

        df_history = get_history_dataframe(self.history, best_fraction=best_fraction)
        columns_needed = ['generation', 'individual', 'fitness']
        df_history = df_history[columns_needed].drop_duplicates(ignore_index=True)
        # Get color palette by mean fitness per generation
        fitness = df_history.groupby('generation')['fitness'].mean()
        fitness = (fitness - min(fitness)) / (max(fitness) - min(fitness))
        colors = plt.cm.YlOrRd(fitness)

        fig, ax = plt.subplots(figsize=(6.4, 4.8), facecolor='w')
        sns.boxplot(data=df_history, x='generation', y='fitness', palette=colors, ax=ax)
        fig.set_dpi(dpi)
        fig.set_facecolor('w')

        ax.set_title('Fitness by generations')
        ax.set_xlabel('Generation')
        # Set ticks for every 5 generation if there's more than 10 generations.
        if len(self.history.generations) > 10:
            ax.xaxis.set_major_locator(ticker.MultipleLocator(5))
            ax.xaxis.set_major_formatter(ticker.ScalarFormatter())
            ax.xaxis.grid(True)
        str_fraction_of_graphs = 'all' if best_fraction is None else f'top {best_fraction * 100}% of'
        ax.set_ylabel(f'Fitness of {str_fraction_of_graphs} generation graphs')
        ax.yaxis.grid(True)

        show_or_save_figure(fig, save_path, dpi)



import datetime
import os
import random
from typing import Type

import numpy as np
import pandas as pd

from golem.core.adapter import DirectAdapter
from golem.core.dag.convert import graph_structure_as_nx_graph
from golem.core.dag.verification_rules import has_no_cycle, has_no_self_cycled_nodes
from golem.core.log import default_log
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.genetic.operators.regularization import RegularizationTypesEnum
from golem.core.optimisers.graph import OptGraph, OptNode
from golem.core.optimisers.objective import Objective, ObjectiveEvaluate
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, GraphOptimizer
from golem.core.optimisers.random.random_mutation_optimizer import RandomMutationOptimizer
from golem.core.optimisers.random.random_search import RandomSearchOptimizer
from golem.core.paths import project_root

random.seed(1)
np.random.seed(1)


class CustomGraphModel(OptGraph):
    def evaluate(self, data: pd.DataFrame):
        nodes = data.columns.to_list()
        _, labels = graph_structure_as_nx_graph(self)
        return len(nodes)


class CustomGraphNode(OptNode):
    def __str__(self):
        return f'Node_{self.name}'


def custom_metric(graph: CustomGraphModel, data: pd.DataFrame, visualisation: bool = False):
    if visualisation:
        graph.show()
    existing_variables_num = -graph.depth - graph.evaluate(data)

    return existing_variables_num


def _has_no_duplicates(graph):
    _, labels = graph_structure_as_nx_graph(graph)
    if len(labels.values()) != len(set(labels.values())):
        raise ValueError('Custom graph has duplicates')
    return True


def custom_mutation(graph: CustomGraphModel, **kwargs) -> CustomGraphModel:
    num_mut = 10
    try:
        for _ in range(num_mut):
            rid = random.choice(range(graph.length))
            random_node = graph.nodes[rid]
            other_random_node = graph.nodes[random.choice(range(len(graph.nodes)))]
            graph.connect_nodes(random_node, other_random_node)
    except Exception as ex:
        default_log(prefix='custom_mutation').warning(f'Incorrect connection: {ex}')
    return graph


def run_custom_example(optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer, timeout: datetime.timedelta = None,
                       visualisation: bool = True):
    if not timeout:
        timeout = datetime.timedelta(minutes=3)

    data = pd.read_csv(os.path.join(project_root(), 'examples', 'data', 'custom_encoded.csv'))
    nodes_types = ['V1', 'V2', 'V3',
                   'V4', 'V5', 'V6',
                   'V7', 'V8', 'V9', 'V10']

    initial = [CustomGraphModel(nodes=[CustomGraphNode(node_type) for node_type in nodes_types])]

    requirements = GraphRequirements(
        max_arity=10,
        max_depth=10,
        timeout=timeout)

    optimizer_parameters = GPAlgorithmParameters(
        pop_size=5,
        crossover_prob=0.8, mutation_prob=0.9,
        genetic_scheme_type=GeneticSchemeTypesEnum.steady_state,
        mutation_types=[custom_mutation],
        crossover_types=[CrossoverTypesEnum.none],
        regularization_type=RegularizationTypesEnum.none)

    adapter = DirectAdapter(base_graph_class=CustomGraphModel, base_node_class=CustomGraphNode)
    constraints = [has_no_self_cycled_nodes, has_no_cycle, _has_no_duplicates]
    graph_generation_params = GraphGenerationParams(
        adapter=adapter,
        rules_for_constraint=constraints,
        available_node_types=nodes_types)

    objective = Objective({'custom': custom_metric})

    optimizer = optimizer_cls(
        objective=objective,
        initial_graphs=initial,
        requirements=requirements,
        graph_generation_params=graph_generation_params,
        graph_optimizer_params=optimizer_parameters)

    objective_eval = ObjectiveEvaluate(objective, data=data, visualisation=visualisation)
    optimized_graphs = optimizer.optimise(objective_eval)
    optimized_network = adapter.restore(optimized_graphs[0])
    if visualisation:
        optimized_network.show()


if __name__ == '__main__':
    visualisation = False
    timeout = datetime.timedelta(minutes=1)
    optimizers = [EvoGraphOptimizer, RandomSearchOptimizer, RandomMutationOptimizer]
    for optimizer_cls in optimizers:
        run_custom_example(optimizer_cls, timeout, visualisation)



import numpy as np

from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.genetic.parameters.parameter import AdaptiveParameter


class AdaptiveMutationProb(AdaptiveParameter[float]):

    def __init__(self, default_prob: float = 0.5):
        self._current_std = 0.
        self._max_std = 0.
        self._min_proba = 0.05
        self._default_prob = default_prob

    @property
    def initial(self) -> float:
        return self._default_prob

    def next(self, population: PopulationT) -> float:
        self._update_std(population)

        if len(population) < 2:
            mutation_prob = 1.0
        elif self._max_std == 0:
            mutation_prob = self._default_prob
        else:
            mutation_prob = max(1. - (self._current_std / self._max_std), self._min_proba)
        return mutation_prob

    def _update_std(self, population: PopulationT):
        self._current_std = self._calc_std(population)
        self._max_std = max(self._current_std, self._max_std)

    @staticmethod
    def _calc_std(population: PopulationT) -> float:
        return float(np.std([ind.fitness.value for ind in population]))



from typing import List, Sequence, Tuple, Optional

import networkx as nx
import numpy as np
from rdkit.Chem.rdchem import Atom, Bond

from examples.molecule_search.mol_graph import MolGraph
from examples.molecule_search.utils import get_default_valence, get_free_electrons_num
from golem.core.optimisers.advisor import DefaultChangeAdvisor


class MolChangeAdvisor(DefaultChangeAdvisor):
    def propose_parent(self, node: Atom, possible_operations: List[str]) -> List[str]:
        """
        Proposes types of atoms that can be connected to the current atom -
        any type other than the type of the current atom.
        """
        atom_types = list(set(possible_operations) - set(node.GetSymbol()))
        return atom_types

    def propose_change(self, node: Atom, possible_operations: List[str]) -> List[str]:
        """
        Proposes types of atoms to replace current atom type - any type other than the type of the current atom
        such that it has enough valence number (valence must be greater or equal to current atom explicit valence).
        """
        atom_types = list(set(possible_operations) - set(node.GetSymbol()))
        atom_types_to_replace = [atom_type for atom_type in atom_types
                                 if get_default_valence(atom_type) >= node.GetExplicitValence()]
        return atom_types_to_replace

    @staticmethod
    def propose_connection_point(mol_graph: MolGraph, bond: Optional[Bond] = None) -> Sequence[int]:
        """
        Proposes atoms new atom or functional group can connect to -
        atoms must have zero formal charge and enough free electrons.
        """
        molecule = mol_graph.get_rw_molecule()
        atoms = molecule.GetAtoms()
        free_electrons_vector = np.array([get_free_electrons_num(atom) for atom in atoms])
        formal_charge_vector = np.array([atom.GetFormalCharge() for atom in atoms])
        atom_ids = np.arange(mol_graph.heavy_atoms_number)

        bond_degree = bond.GetBondTypeAsDouble() if bond else 1  # BondType.SINGLE is used by default
        return list(atom_ids[(formal_charge_vector == 0) & (free_electrons_vector >= bond_degree)])

    @staticmethod
    def propose_atom_removal(mol_graph: MolGraph) -> Sequence[int]:
        """
        Proposes atoms that can be removed - any atom, which deletion will not increase the number of connected
        components of the molecule.
        """
        if mol_graph.heavy_atoms_number > 1:
            nx_graph = mol_graph.get_nx_graph()
            art_points_ids = nx.articulation_points(nx_graph)
            atom_ids = np.arange(mol_graph.heavy_atoms_number)
            return list(set(atom_ids) - set(art_points_ids))
        else:
            return []

    @staticmethod
    def propose_connection(mol_graph: MolGraph) -> Sequence[Tuple[int, int]]:
        """
        Proposes pairs of atoms that can be connected - both atoms must have zero formal charge.
        """
        molecule = mol_graph.get_rw_molecule()
        atoms = np.array(molecule.GetAtoms())
        atom_pairs_to_connect = []
        for from_atom in atoms:
            for to_atom in atoms[from_atom.GetIdx() + 1:]:
                can_be_connected = to_atom.GetFormalCharge() == 0 and from_atom.GetFormalCharge() == 0
                if can_be_connected:
                    atom_pairs_to_connect.append((from_atom.GetIdx(), to_atom.GetIdx()))
        return atom_pairs_to_connect

    @staticmethod
    def propose_disconnection(mol_graph: MolGraph) -> Sequence[Tuple[int, int]]:
        """
        Proposes pairs of atoms that can be disconnected - both atoms must have zero formal charge and deletion
        of the bond must not increase the number of connected components of the molecule.
        """
        bridges = set(nx.bridges(mol_graph.get_nx_graph()))
        atom_pairs_to_disconnect = []
        for bond in mol_graph.get_rw_molecule().GetBonds():
            from_atom = bond.GetBeginAtom()
            to_atom = bond.GetEndAtom()
            can_be_disconnected = to_atom.GetFormalCharge() == 0 and from_atom.GetFormalCharge() == 0
            if can_be_disconnected:
                atom_pairs_to_disconnect.append((from_atom.GetIdx(), to_atom.GetIdx()))
        atom_pairs_to_disconnect = set(atom_pairs_to_disconnect) - bridges
        return list(atom_pairs_to_disconnect)

    @staticmethod
    def propose_cut(mol_graph: MolGraph) -> Sequence[int]:
        """
        Proposes atoms that can be cut - any atom that has only two neighbors, the neighbors are disconnected,
        and they have zero formal charge.
        """
        molecule = mol_graph.get_rw_molecule()
        atoms_to_cut = []
        for atom in molecule.GetAtoms():
            neighbors = atom.GetNeighbors()
            if len(neighbors) == 2:
                if (not molecule.GetBondBetweenAtoms(neighbors[0].GetIdx(), neighbors[1].GetIdx()) and
                        neighbors[0].GetFormalCharge() == 0 and
                        neighbors[1].GetFormalCharge() == 0):
                    atoms_to_cut.append(atom.GetIdx())
        return atoms_to_cut

    @staticmethod
    def propose_bond_to_split(mol_graph: MolGraph) -> Sequence[Tuple[int, int]]:
        """
        Proposes bonds which can be split by new atom - bond endpoints have no formal charge.
        """
        bonds_to_split = []
        for bond in mol_graph.get_rw_molecule().GetBonds():
            from_atom = bond.GetBeginAtom()
            to_atom = bond.GetEndAtom()
            if from_atom.GetFormalCharge() == 0 and to_atom.GetFormalCharge() == 0:
                bonds_to_split.append((from_atom.GetIdx(), to_atom.GetIdx()))
        return bonds_to_split

    @staticmethod
    def propose_group(mol_graph: MolGraph) -> List[Tuple[int, int]]:
        """
        Proposes functional groups that can be operated. Group is a connected subgraph that is connected
        to the rest of the graph by only one bond.

        Returns:
            List of indices pairs which are endpoints of bridges. This bridges connect functional group to the molecule.
            The second index of each pair belongs to the functional group.
        """
        molecule = mol_graph.get_rw_molecule()
        nx_graph = mol_graph.get_nx_graph()
        groups = []
        for bridge in nx.bridges(nx_graph):
            from_atom = molecule.GetAtomWithIdx(bridge[0])
            to_atom = molecule.GetAtomWithIdx(bridge[1])
            can_be_disconnected = to_atom.GetFormalCharge() == 0 and from_atom.GetFormalCharge() == 0
            if can_be_disconnected:
                groups.extend([bridge, bridge[::-1]])
        return groups



import collections.abc
import dataclasses

from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from typing import Callable, Container, Generic, Iterable, Iterator, List, Optional, Sequence, Sized, TypeVar, Union, \
    Tuple, Any, Dict

T = TypeVar('T')


class UniqueList(list, Generic[T]):
    """
    Simple list that maintains uniqueness of elements.
    In comparison to 'set': preserves list interface and element ordering.

    But this class is better to use only for short lists because of linear complexity of uniqueness lookup.
    Behaves optimally for <= 4 items, good enough for <= 20 items.

    List addition and multiplication are not overloaded and return a standard list.
    But addition with assignment (+=) behaves as UniqueList.extend.

    :param iterable: list of elements to turn into distinct
    """

    def __init__(self, iterable: Optional[Iterable[T]] = None):
        iterable = iterable or ()
        super().__init__(dict.fromkeys(iterable).keys())  # preserves order and uniqueness in the newly created list

    def append(self, value: T):
        """
        Adds ``value`` to the end of the list in case it is not present there

        :param value: will be added to the list or ignored if it's already there
        """
        if value not in super().__iter__():
            super().append(value)

    def extend(self, iterable: Iterable[T]):
        """
        Extends current list with the ``iterable`` elements that aren't present in the list yet

        :param iterable: sequence of elements to be added to the list in case they aren't there
        """
        impl = super()
        impl.extend(element for element in iterable
                    if element not in impl.__iter__())

    def insert(self, index: int, value: T):
        """
        Inserts specified ``value`` to the provided ``index`` in the list in case it's unique value

        :param index: position for inserting into the list
        :param value: will be inserted at the specified ``index`` if it's unique
        """
        if value not in super().__iter__():
            super().insert(index, value)

    def __setitem__(self, key: Union[int, slice], value: Union[T, Iterable[T]]):
        """
        Sets specified ``value`` at the specified index or slice if it's unique value

        :param key: for pointing to the elements from the list
        :param value: element(s) to be set by the specified ``key``
        """
        if value not in super().__iter__():
            super().__setitem__(key, value)

    def __iadd__(self, other: Iterable[T]) -> 'UniqueList':
        """
        Extends current list with the ``iterable`` elements that aren't present in the list yet

        :param other: sequence of elements to be added to the list in case they aren't there

        :return: this class instance
        """
        self.extend(other)
        return self


def remove_items(collection: List[T], items_to_remove: Container[T]):
    """
    Removes all specified items from the list. Modifies original collection

    :param collection: list of elements to be filtered
    :param items_to_remove: filter list of unwanted elements

    :return: modified ``collection`` parameter
    """
    if collection:
        collection[:] = [item for item in collection if item not in items_to_remove]
    return collection


def are_same_length(collections: Iterable[Sized]) -> bool:
    """
    Checks if all arguments have the same length

    :param collections: collection of collections

    :return: does collections inside ``collections`` have the same length
    """
    it = collections.__iter__()
    first = next(it, None)
    if first is not None:
        first = len(first)
        for elem in it:
            if len(elem) != first:
                return False
    return True


def ensure_wrapped_in_sequence(
        obj: Optional[Union[T, Iterable[T]]],
        sequence_factory: Callable[[Iterable[T]], Sequence[T]] = list
) -> Optional[Sequence[T]]:
    """
    Makes sure given ``obj`` is of type that acts like sequence and converts ``obj`` to it otherwise

    :param obj: any object to be ensured or wrapped into sequence type
    :param sequence_factory: sequence factory for wrapping ``obj`` into in case it is not of any sequence type

    :return: the same object if it's of type sequence (or None)
        or ``obj`` wrapped in type provided by ``sequence_factory``
    """
    if obj is None:
        return obj
    elif isinstance(obj, str) or not isinstance(obj, collections.abc.Iterable):
        return sequence_factory([obj])
    elif not isinstance(obj, collections.abc.Sequence) or not isinstance(obj, sequence_factory):
        return sequence_factory(obj)
    return obj


class ComparableEnum(Enum):
    """
    The Enum implementation that allows to avoid the multi-module enum comparison problem
    (https://stackoverflow.com/questions/26589805/python-enums-across-modules)
    """

    def __eq__(self, other: 'Enum'):
        """
        Compares this enum with the ``other_graph``

        :param other: another enum

        :return: is it equal to ``other`` in terms of the string representation
        """
        return str(self) == str(other)

    def __hash__(self):
        """
        Gets hashcode of this enum

        :return: hashcode of string representation of this enum
        """
        return hash(str(self))


class Copyable:
    """Provides default implementations for `copy` & `deepcopy`."""

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo=None):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result


class Comparable(ABC):
    @abstractmethod
    def __eq__(self, other: 'Comparable') -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is it equal to ``other`` in terms of the comparables
        """
        raise NotImplementedError()

    @abstractmethod
    def __lt__(self, other: 'Comparable') -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is less than ``other`` in terms of the comparables
        """
        raise NotImplementedError()

    def __ne__(self, other: 'Comparable') -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is it not equal to ``other`` in terms of the comparables
        """
        return not self.__eq__(other)

    def __le__(self, other) -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is less than or equal to ``other`` in terms of the comparables
        """
        return self.__lt__(other) or self.__eq__(other)

    def __gt__(self, other) -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is greater than ``other`` in terms of the comparables
        """
        return not self.__le__(other)

    def __ge__(self, other) -> bool:
        """
        Compares this object of :class:`Comparable` interface with the ``other`` comparable

        :param other: other object implementing :class:`Comparable` interface

        :return: is greater than or equal to ``other`` in terms of the comparables
        """
        return not self.__lt__(other)


class BidirectionalIterator(Iterator[T]):
    @abstractmethod
    def has_prev(self) -> bool:
        """
        Checks if this iterator has implemented previous item getter

        :return: whether this iterator has previous item getter
        """
        raise NotImplementedError()

    @abstractmethod
    def has_next(self) -> bool:
        """
        Checks if this iterator has implemented next item getter

        :return: whether this iterator has next item getter
        """
        raise NotImplementedError()

    @abstractmethod
    def next(self) -> T:
        """
        Gets next item of this iterator

        :return: next item
        """
        raise NotImplementedError()

    @abstractmethod
    def prev(self) -> T:
        """
        Gets previous item of this iterator

        :return: previous item
        """
        raise NotImplementedError()

    @abstractmethod
    def current(self) -> T:
        """
        Get current value pointed to by the iterator

        :return: current item
        """
        raise NotImplementedError()

    def __next__(self):
        """
        Gets next item of this iterator

        :raises StopIteration: the end of this iterator

        :return: next item
        """
        if self.has_next():
            return self.next()
        raise StopIteration()

    def __iter__(self):
        """
        Returns this iterator

        :return: this iterator
        """
        return self


def unzip(tuples: Iterable[Tuple]) -> Tuple[Sequence, ...]:
    return tuple(zip(*tuples))


def update_dataclass(base_dc: Any, update_dc: Union[Any, Dict]) -> Any:
    update_dict = dataclasses.asdict(update_dc) if dataclasses.is_dataclass(update_dc) else update_dc
    new_base = dataclasses.replace(base_dc, **update_dict)
    return new_base



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from golem.core.log import default_log
from golem.visualisation.opt_history.diversity import DiversityLine, DiversityPopulation
from golem.visualisation.opt_history.fitness_box import FitnessBox
from golem.visualisation.opt_history.fitness_line import FitnessLine, FitnessLineInteractive
from golem.visualisation.opt_history.operations_animated_bar import OperationsAnimatedBar
from golem.visualisation.opt_history.operations_kde import OperationsKDE

if TYPE_CHECKING:
    from golem.core.optimisers.opt_history_objects.opt_history import OptHistory


class PlotTypesEnum(Enum):
    fitness_box = FitnessBox
    fitness_line = FitnessLine
    fitness_line_interactive = FitnessLineInteractive
    operations_kde = OperationsKDE
    operations_animated_bar = OperationsAnimatedBar
    diversity_line = DiversityLine
    diversity_population = DiversityPopulation

    @classmethod
    def member_names(cls):
        return cls._member_names_


class OptHistoryVisualizer:
    """ Implements optimization history visualizations available via `OptHistory.show()`.

    `OptHistory.show` points to the initialized instance of this class.
    Thus, supported visualizations can be called directly via `OptHistory.show.<vis_name>(...)` or
    indirectly via `OptHistory.show(...)`.

    This implies that all supported visualizations are listed in two places:
        1. `PlotTypesEnum` members

        `<vis_name> = <VisClass>`

        2. assigned as the class attributes in `OptHistoryVisualizer.__init__`

        `self.<vis_name> = <VizClass>(self.history).visualize`
    """

    def __init__(self, history: OptHistory, visuals_params: Optional[Dict[str, Any]] = None):
        visuals_params = visuals_params or {}
        default_visuals_params = dict(dpi=100)
        default_visuals_params.update(visuals_params)

        self.visuals_params = default_visuals_params
        self.history = history

        self.fitness_box = FitnessBox(history, self.visuals_params).visualize
        self.fitness_line = FitnessLine(history, self.visuals_params).visualize
        self.fitness_line_interactive = FitnessLineInteractive(history, self.visuals_params).visualize
        self.operations_kde = OperationsKDE(history, self.visuals_params).visualize
        self.operations_animated_bar = OperationsAnimatedBar(history, self.visuals_params).visualize
        self.diversity_line = DiversityLine(history, self.visuals_params).visualize
        self.diversity_population = DiversityPopulation(history, self.visuals_params).visualize

        self.log = default_log(self)

    def __call__(self, plot_type: Union[PlotTypesEnum, str] = PlotTypesEnum.fitness_line, **kwargs):
        """ Visualizes the OptHistory with one of the supported methods.

        :param plot_type: visualization to show. Expected values are listed in
            'golem.core.visualisation.opt_viz.PlotTypesEnum'.
        :keyword save_path: path to save the visualization. If set, then the image will be saved, and if not,
            it will be displayed. Essential for animations.
        :keyword dpi: DPI of the output figure.
        :keyword best_fraction: fraction of the best individuals of each generation that included in the
            visualization. Must be in the interval (0, 1].
        :keyword show_fitness: if False, visualizations that support this argument will not display fitness.
        :keyword per_time: Shows time axis instead of generations axis. Currently, supported for plot types:
            'show_fitness_line', 'show_fitness_line_interactive'.
        :keyword use_tags: if True (default), all operations in the history are colored and grouped based on
            operation repository tags. If False, operations are not grouped, colors are picked by fixed colormap
            for every history independently.
        """

        if isinstance(plot_type, str):
            try:
                visualize_function = vars(self)[plot_type]
            except KeyError:
                raise NotImplementedError(
                    f'Visualization "{plot_type}" is not supported. Expected values: '
                    f'{", ".join(PlotTypesEnum.member_names())}.')
        else:
            visualize_function = vars(self)[plot_type.name]
        visualize_function(**kwargs)



import random
from collections import namedtuple
from typing import List, Optional

from golem.core.dag.graph_verifier import GraphVerifier
from golem.core.dag.verification_rules import DEFAULT_DAG_RULES
from golem.core.dag.graph import GraphNode
from golem.structural_analysis.graph_sa.entities.edge import Edge

ReplacementAnalysisMetaParams = namedtuple('ReplacementAnalysisMetaParams', ['edges_to_replace_to',
                                                                             'number_of_random_operations_edges',
                                                                             'nodes_to_replace_to',
                                                                             'number_of_random_operations_nodes'])


class StructuralAnalysisRequirements:
    """
    Use this object to pass all the requirements needed for SA.

    :param graph_verifier: verifier for graph in SA.
    :param main_metric_idx: index of metric to optimize by. Other metrics will be calculated and saved if needed.
    :param replacement_nodes_to_replace_to: defines nodes which is used in replacement analysis
    :param replacement_number_of_random_operations_nodes: if replacement_nodes_to_replace_to is not filled, \
    define the number of randomly chosen operations used in replacement analysis
    :param replacement_edges_to_replace_to: defines edges which is used in replacement analysis
    :param replacement_number_of_random_operations_edges: if replacement_edges_to_replace_to is not filled, \
    define the number of randomly chosen operations used in replacement analysis
    :param is_visualize: defines whether the SA visualization needs to be saved to .png files
    :param is_save_results_to_json: defines whether the SA indices needs to be saved to .json file
    """

    def __init__(self,
                 graph_verifier: GraphVerifier = None,
                 main_metric_idx: int = 0,
                 replacement_nodes_to_replace_to: Optional[List[GraphNode]] = None,
                 replacement_number_of_random_operations_nodes: Optional[int] = 3,
                 replacement_edges_to_replace_to: Optional[List[Edge]] = None,
                 replacement_number_of_random_operations_edges: Optional[int] = 3,
                 is_visualize: bool = False,
                 is_save_results_to_json: bool = False,
                 seed: int = random.randint(0, 100)):

        self.graph_verifier = graph_verifier or GraphVerifier(DEFAULT_DAG_RULES)

        self.main_metric_idx = main_metric_idx

        self.replacement_meta = ReplacementAnalysisMetaParams(replacement_edges_to_replace_to,
                                                              replacement_number_of_random_operations_edges,
                                                              replacement_nodes_to_replace_to,
                                                              replacement_number_of_random_operations_nodes)

        self.is_visualize = is_visualize
        self.is_save = is_save_results_to_json
        self.seed = seed



from typing import TYPE_CHECKING

from golem.core.optimisers.genetic.operators.operator import PopulationT, Operator
from golem.core.optimisers.genetic.operators.selection import Selection
from golem.utilities.data_structures import ComparableEnum as Enum

if TYPE_CHECKING:
    from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters


class GeneticSchemeTypesEnum(Enum):
    steady_state = 'steady_state'
    generational = 'generational'
    parameter_free = 'parameter_free'


class Inheritance(Operator):
    def __init__(self, parameters: 'GPAlgorithmParameters', selection: Selection):
        super().__init__(parameters=parameters)
        self.selection = selection

    def __call__(self, previous_population: PopulationT, new_population: PopulationT) -> PopulationT:
        gp_scheme = self.parameters.genetic_scheme_type
        if gp_scheme == GeneticSchemeTypesEnum.generational:
            # Previous population is completely substituted
            next_population = self.direct_inheritance(new_population)
        elif gp_scheme == GeneticSchemeTypesEnum.steady_state:
            # Previous population is mixed with new one
            next_population = self.steady_state_inheritance(previous_population, new_population)
        elif gp_scheme == GeneticSchemeTypesEnum.parameter_free:
            # Same as steady-state
            next_population = self.steady_state_inheritance(previous_population, new_population)
        else:
            raise ValueError(f'Unknown genetic scheme {gp_scheme}!')
        return next_population

    def steady_state_inheritance(self,
                                 prev_population: PopulationT,
                                 new_population: PopulationT
                                 ) -> PopulationT:
        # use individuals with non-repetitive uid
        not_repetitive_inds = [ind for ind in prev_population if ind not in new_population]
        full_population = new_population + not_repetitive_inds
        return self.selection(full_population,
                              pop_size=self.parameters.pop_size)

    def direct_inheritance(self, new_population: PopulationT) -> PopulationT:
        return new_population[:self.parameters.pop_size]



from typing import Callable, Sequence, Tuple

import networkx as nx
import numpy as np

from golem.metrics import mmd
from golem.metrics.mmd import compute_mmd


def compute_all_stats(graph_prediction: Sequence[nx.Graph],
                      graph_target: Sequence[nx.Graph]) -> Tuple[float, float]:
    mmd_degree = degree_stats(graph_prediction, graph_target)
    mmd_clustering = clustering_stats(graph_prediction, graph_target)

    return mmd_degree, mmd_clustering


def degree_stats(graph_prediction: Sequence[nx.Graph],
                 graph_target: Sequence[nx.Graph]) -> float:
    return mmd_stats(nx.degree_histogram, graph_prediction, graph_target, normalize=True)


def clustering_stats(graph_prediction: Sequence[nx.Graph],
                     graph_target: Sequence[nx.Graph]) -> float:
    bins = 100
    return mmd_stats(clustering_stats_graph, graph_prediction, graph_target,
                     sigma=0.1, distance_scaling=bins, normalize=False)


def mmd_stats(stat_function: Callable[[nx.Graph], np.ndarray],
              graph_prediction: Sequence[nx.Graph],
              graph_target: Sequence[nx.Graph],
              **kwargs):
    sample_predict = list(map(stat_function, graph_prediction))
    sample_target = list(map(stat_function, graph_target))
    return compute_mmd(sample_target, sample_predict, **kwargs)


def mmd_stats_impl(stat_function: Callable[[nx.Graph], np.ndarray],
                   graph_prediction: Sequence[nx.Graph],
                   graph_target: Sequence[nx.Graph],
                   kernel: Callable = mmd.gaussian_emd,
                   sigma: float = 1.0,
                   distance_scaling: float = 1.0,
                   normalize: bool = False) -> float:

    sample_predict = list(map(stat_function, graph_prediction))
    sample_target = list(map(stat_function, graph_target))

    mmd_dist = compute_mmd(sample_target, sample_predict,
                           normalize=normalize, kernel=kernel,
                           sigma=sigma, distance_scaling=distance_scaling)
    return mmd_dist


def clustering_stats_graph(graph: nx.Graph, bins: int = 100) -> np.ndarray:
    clustering_coeffs = list(nx.clustering(graph).values())
    hist, _ = np.histogram(clustering_coeffs, bins=bins, range=(0.0, 1.0), density=True)
    return hist



# flake8: noqa
from .enum_serialization import enum_from_json, enum_to_json
from .graph_node_serialization import graph_node_to_json
from .graph_serialization import graph_from_json
from .opt_history_serialization import opt_history_from_json, opt_history_to_json
from .parent_operator_serialization import parent_operator_from_json, parent_operator_to_json
from .uuid_serialization import uuid_from_json, uuid_to_json



import math
from typing import Optional

from golem.core.constants import MIN_POP_SIZE
from golem.core.optimisers.archive.generation_keeper import ImprovementWatcher
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.genetic.parameters.parameter import AdaptiveParameter
from golem.utilities.data_structures import BidirectionalIterator
from golem.utilities.sequence_iterator import fibonacci_sequence, SequenceIterator

PopulationSize = AdaptiveParameter[int]


class ConstRatePopulationSize(PopulationSize):
    def __init__(self, pop_size: int, offspring_rate: float, max_pop_size: Optional[int] = None):
        self._offspring_rate = offspring_rate
        self._initial = pop_size
        self._max_pop_size = max_pop_size

    @property
    def initial(self) -> int:
        return self._initial

    def next(self, population: PopulationT) -> int:
        # to prevent stagnation
        pop_size = max(len(population), self._initial)
        if not self._max_pop_size or pop_size < self._max_pop_size:
            pop_size += math.ceil(pop_size * self._offspring_rate)
        if self._max_pop_size:
            pop_size = min(pop_size, self._max_pop_size)
        return pop_size


class AdaptivePopulationSize(PopulationSize):
    def __init__(self,
                 improvement_watcher: ImprovementWatcher,
                 progression_iterator: BidirectionalIterator[int],
                 max_pop_size: Optional[int] = None):
        self._improvements = improvement_watcher
        self._iterator = progression_iterator
        self._max_pop_size = max_pop_size
        self._initial = self._iterator.next() if self._iterator.has_next() else self._iterator.prev()

    @property
    def initial(self) -> int:
        return self._initial

    def next(self, population: PopulationT) -> int:
        pop_size = len(population)
        too_many_fitness_eval_errors = pop_size / self._iterator.current() < 0.5

        if too_many_fitness_eval_errors or not self._improvements.is_any_improved:
            if self._iterator.has_next():
                pop_size = self._iterator.next()
        elif self._improvements.is_quality_improved and self._improvements.is_complexity_improved and pop_size > 0:
            if self._iterator.has_prev():
                pop_size = self._iterator.prev()

        pop_size = max(pop_size, MIN_POP_SIZE)
        if self._max_pop_size:
            pop_size = min(pop_size, self._max_pop_size)

        return pop_size


def init_adaptive_pop_size(requirements: GPAlgorithmParameters,
                           improvement_watcher: ImprovementWatcher) -> PopulationSize:
    genetic_scheme_type = requirements.genetic_scheme_type
    if genetic_scheme_type == GeneticSchemeTypesEnum.steady_state:
        pop_size = ConstRatePopulationSize(
            pop_size=requirements.pop_size,
            offspring_rate=1.0,
            max_pop_size=requirements.max_pop_size,
        )
    elif genetic_scheme_type == GeneticSchemeTypesEnum.generational:
        pop_size = ConstRatePopulationSize(
            pop_size=requirements.pop_size,
            offspring_rate=requirements.offspring_rate,
            max_pop_size=requirements.max_pop_size,
        )
    elif genetic_scheme_type == GeneticSchemeTypesEnum.parameter_free:
        pop_size_progression = SequenceIterator(sequence_func=fibonacci_sequence,
                                                start_value=requirements.pop_size,
                                                min_sequence_value=1,
                                                max_sequence_value=requirements.max_pop_size)
        pop_size = AdaptivePopulationSize(improvement_watcher=improvement_watcher,
                                          progression_iterator=pop_size_progression,
                                          max_pop_size=requirements.max_pop_size)
    else:
        raise ValueError(f"Unknown genetic type scheme {genetic_scheme_type}")
    return pop_size



from typing import List, Tuple


class FitnessRateRankRewardTransformer:
    """
    Agent to process raw fitness values.
    The original idea is that the use of raw fitness values as rewards affects algorithm's robustness,
    therefore fitness values must be processed.
    The article with explanation -- https://ieeexplore.ieee.org/document/6410018
    """
    def __init__(self, decaying_factor: float = 1.):
        self._decaying_factor = decaying_factor

    def get_rewards_for_arms(self, rewards: List[float], arms: List[int]) -> List[float]:
        unique_arms, decay_values = self.get_decay_values_for_arms(rewards, arms)
        frr_per_arm = self.get_fitness_rank_rate(decay_values)
        frr_values = [frr_per_arm[unique_arms.index(arm)] for arm in arms]
        return frr_values

    def get_decay_values_for_arms(self, rewards: List[float], arms: List[int]) -> Tuple[List[int], List[float]]:
        decays = dict.fromkeys(set(arms), 0.0)
        for i, reward in enumerate(rewards):
            decays[arms[i]] += reward
        decays.update((key, value * self._decaying_factor) for key, value in decays.items())
        return list(decays.keys()), list(decays.values())

    @staticmethod
    def get_fitness_rank_rate(decay_values: List[float]) -> List[float]:
        # abs() is used to save the initial sign of each decay value
        total_decay_sum = abs(sum(decay_values))
        return [decay / total_decay_sum for decay in decay_values] if total_decay_sum != 0 else [0.]*len(decay_values)



import datetime
from abc import ABC

from golem.core.log import default_log


class Timer(ABC):
    def __init__(self, timeout: datetime.timedelta = None):
        self.process_terminated = False
        self.log = default_log(self)
        self.timeout = timeout

    def __enter__(self):
        self.start = datetime.datetime.now()
        return self

    @property
    def start_time(self):
        return self.start

    @property
    def spent_time(self) -> datetime.timedelta:
        return datetime.datetime.now() - self.start_time

    @property
    def minutes_from_start(self) -> float:
        return self.spent_time.total_seconds() / 60.

    @property
    def seconds_from_start(self) -> float:
        return self.spent_time.total_seconds()

    def is_time_limit_reached(self) -> bool:
        self.process_terminated = False
        if self.timeout is not None:
            if datetime.datetime.now() - self.start >= self.timeout:
                self.process_terminated = True
        return self.process_terminated

    def __exit__(self, *args):
        return self.process_terminated


class OptimisationTimer(Timer):
    def __init__(self, timeout: datetime.timedelta = None):
        super().__init__(timeout=timeout)
        self.init_time = 0

    def _is_next_iteration_possible(self, time_constraint: float, iteration_num: int = None) -> bool:
        minutes = self.minutes_from_start
        if iteration_num is not None and iteration_num != 0:
            evo_proc_minutes = minutes - self.init_time
            possible = time_constraint > (minutes + (evo_proc_minutes / iteration_num))
        else:
            possible = time_constraint > minutes
        if not possible:
            self.process_terminated = True
        return possible

    def is_time_limit_reached(self, iteration_num: int = None) -> bool:
        if self.timeout:
            timeout = 0 if self.timeout.total_seconds() < 0 else self.timeout.total_seconds() / 60.
            if timeout:
                reached = not self._is_next_iteration_possible(iteration_num=iteration_num,
                                                               time_constraint=timeout)
            else:
                self.process_terminated = True
                reached = True
        else:
            reached = False
        return reached

    def set_init_time(self, init_time: float):
        self.init_time = init_time

    def __exit__(self, *args):
        self.log.info(f'Composition time: {round(self.minutes_from_start, 3)} min')
        if self.process_terminated:
            self.log.info('Algorithm was terminated due to processing time limit')


def get_forever_timer() -> Timer:
    return Timer(timeout=None)



from threading import RLock


class SingletonMeta(type):
    """
    This meta class can provide other classes with the Singleton pattern.
    It guarantees to create one and only class instance.
    Pass it to the metaclass parameter when defining your class as follows:

    class YourClassName(metaclass=SingletonMeta)
    """
    _instances = {}
    _lock: RLock = RLock()  # TODO: seems like it's useless in multiprocessing, but that's even from threading lib?!

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]



import networkx as nx
from networkx import isolates

from golem.core.adapter import register_native
from golem.core.dag.convert import graph_structure_as_nx_graph
from golem.core.dag.graph import Graph
from golem.core.dag.graph_node import GraphNode
from golem.core.dag.graph_utils import graph_has_cycle

ERROR_PREFIX = 'Invalid graph configuration:'


@register_native
def has_root(graph: Graph):
    if graph.root_node:
        return True


@register_native
def has_one_root(graph: Graph):
    if isinstance(graph.root_node, GraphNode):
        return True


@register_native
def has_no_cycle(graph: Graph):
    if graph_has_cycle(graph):
        raise ValueError(f'{ERROR_PREFIX} Graph has cycles')

    return True


@register_native
def has_no_isolated_nodes(graph: Graph):
    nx_graph, _ = graph_structure_as_nx_graph(graph)
    isolated = list(isolates(nx_graph))
    if len(isolated) > 0 and graph.length != 1:
        raise ValueError(f'{ERROR_PREFIX} Graph has isolated nodes')
    return True


@register_native
def has_no_self_cycled_nodes(graph: Graph):
    if any([node for node in graph.nodes if node.nodes_from and node in node.nodes_from]):
        raise ValueError(f'{ERROR_PREFIX} Graph has self-cycled nodes')
    return True


@register_native
def has_no_isolated_components(graph: Graph):
    nx_graph, _ = graph_structure_as_nx_graph(graph)
    ud_nx_graph = nx.Graph()
    ud_nx_graph.add_nodes_from(nx_graph)
    ud_nx_graph.add_edges_from(nx_graph.edges)
    if ud_nx_graph.number_of_nodes() == 0:
        raise ValueError(f'{ERROR_PREFIX} Graph is null, connectivity not defined')
    if not nx.is_connected(ud_nx_graph):
        raise ValueError(f'{ERROR_PREFIX} Graph has isolated components')
    return True


DEFAULT_DAG_RULES = [has_root, has_no_cycle, has_no_isolated_components,
                     has_no_self_cycled_nodes, has_no_isolated_nodes]



from functools import partial
from typing import Tuple, Optional

from hyperopt import Trials, fmin, space_eval

from golem.core.constants import MIN_TIME_FOR_TUNING_IN_SEC
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.timer import Timer
from golem.core.tuning.hyperopt_tuner import HyperoptTuner, get_node_parameters_for_hyperopt
from golem.core.tuning.search_space import get_node_operation_parameter_label
from golem.core.tuning.tuner_interface import DomainGraphForTune


class SimultaneousTuner(HyperoptTuner):
    """
        Class for hyperparameters optimization for all nodes simultaneously
    """

    def tune(self, graph: DomainGraphForTune, show_progress: bool = True) -> DomainGraphForTune:
        """ Function for hyperparameters tuning on the entire graph

        Args:
            graph: graph which hyperparameters will be tuned
            show_progress: shows progress of tuning if True

        Returns:
            Graph with tuned hyperparameters
        """

        graph = self.adapter.adapt(graph)
        parameters_dict, init_parameters = self._get_parameters_for_tune(graph)

        with Timer() as global_tuner_timer:
            self.init_check(graph)
            self._update_remaining_time(global_tuner_timer)

            if not parameters_dict:
                self._stop_tuning_with_message(f'Graph "{graph.graph_description}" has no parameters to optimize')
                final_graph = graph

            elif self.max_seconds <= MIN_TIME_FOR_TUNING_IN_SEC:
                self._stop_tuning_with_message('Tunner stopped after initial assumption due to the lack of time')
                final_graph = graph

            else:
                trials = Trials()

                try:
                    # try searching using initial parameters
                    # (uses original search space with fixed initial parameters)
                    trials, init_trials_num = self._search_near_initial_parameters(graph,
                                                                                   parameters_dict,
                                                                                   init_parameters,
                                                                                   trials,
                                                                                   show_progress)
                    self._update_remaining_time(global_tuner_timer)
                    if self.max_seconds > MIN_TIME_FOR_TUNING_IN_SEC:
                        fmin(partial(self._objective, graph=graph),
                             parameters_dict,
                             trials=trials,
                             algo=self.algo,
                             max_evals=self.iterations,
                             show_progressbar=show_progress,
                             early_stop_fn=self.early_stop_fn,
                             timeout=self.max_seconds)
                    else:
                        self.log.message('Tunner stopped after initial search due to the lack of time')

                    best = space_eval(space=parameters_dict, hp_assignment=trials.argmin)
                    # check if best point was obtained using search space with fixed initial parameters
                    is_best_trial_with_init_params = trials.best_trial.get('tid') in range(init_trials_num)
                    if is_best_trial_with_init_params:
                        best = {**best, **init_parameters}

                    final_graph = self.set_arg_graph(graph=graph, parameters=best)

                    self.was_tuned = True

                except Exception as ex:
                    self.log.warning(f'Exception {ex} occurred during tuning')
                    final_graph = graph

        # Validate if optimisation did well
        graph = self.final_check(final_graph)
        final_graph = self.adapter.restore(graph)
        return final_graph

    def _search_near_initial_parameters(self, graph: OptGraph,
                                        search_space: dict,
                                        initial_parameters: dict,
                                        trials: Trials,
                                        show_progress: bool = True) -> Tuple[Trials, int]:
        """ Method to search using the search space where parameters initially set for the graph are fixed.
        This allows not to lose results obtained while composition process

        Args:
            graph: graph to be tuned
            search_space: dict with parameters to be optimized and their search spaces
            initial_parameters: dict with initial parameters of the graph
            trials: Trials object to store all the search iterations
            show_progress: shows progress of tuning if True

        Returns:
            trials: Trials object storing all the search trials
            init_trials_num: number of iterations made using the search space with fixed initial parameters
        """
        try_initial_parameters = initial_parameters and self.iterations > 1
        if not try_initial_parameters:
            init_trials_num = 0
            return trials, init_trials_num

        is_init_params_full = len(initial_parameters) == len(search_space)
        if self.iterations < 10 or is_init_params_full:
            init_trials_num = 1
        else:
            init_trials_num = min(int(self.iterations * 0.1), 10)

        # fmin updates trials with evaluation points tried out during the call
        fmin(partial(self._objective, graph=graph, unchangeable_parameters=initial_parameters),
             search_space,
             trials=trials,
             algo=self.algo,
             max_evals=init_trials_num,
             show_progressbar=show_progress,
             early_stop_fn=self.early_stop_fn,
             timeout=self.max_seconds)
        return trials, init_trials_num

    def _get_parameters_for_tune(self, graph: OptGraph) -> Tuple[dict, dict]:
        """ Method for defining the search space

        Args:
            graph: graph to be tuned

        Returns:
            parameters_dict: dict with operation names and parameters
            initial_parameters: dict with initial parameters of the graph
        """

        parameters_dict = {}
        initial_parameters = {}
        for node_id, node in enumerate(graph.nodes):
            operation_name = node.name

            # Assign unique prefix for each model hyperparameter
            # label - number of node in the graph
            node_params = get_node_parameters_for_hyperopt(self.search_space, node_id=node_id,
                                                           operation_name=operation_name)
            parameters_dict.update(node_params)

            tunable_node_params = self.search_space.get_parameters_for_operation(operation_name)
            if tunable_node_params:
                tunable_initial_params = {get_node_operation_parameter_label(node_id, operation_name, p):
                                          node.parameters[p] for p in node.parameters if p in tunable_node_params}
                if tunable_initial_params:
                    initial_parameters.update(tunable_initial_params)

        return parameters_dict, initial_parameters

    def _objective(self, parameters_dict: dict, graph: OptGraph, unchangeable_parameters: Optional[dict] = None) \
            -> float:
        """
        Objective function for minimization problem

        Args:
            parameters_dict: dict which contains new graph hyperparameters
            graph: graph to optimize
            unchangeable_parameters: dict with parameters that should not be changed

        Returns:
             metric_value: value of objective function
        """

        # replace new parameters with parameters
        if unchangeable_parameters:
            parameters_dict = {**parameters_dict, **unchangeable_parameters}

        # Set hyperparameters for every node
        graph = self.set_arg_graph(graph, parameters_dict)

        metric_value = self.get_metric_value(graph=graph)

        return metric_value



from random import choice

from golem.core.optimisers.genetic.operators.base_mutations import single_drop_mutation, add_as_child, \
    add_separate_parent_node, add_intermediate_node
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, AlgorithmParameters


# mutations which rewards are changing during the optimization process


def fake_add_mutation(graph: OptGraph,
                      requirements: GraphRequirements,
                      graph_gen_params: GraphGenerationParams,
                      parameters: AlgorithmParameters) -> OptGraph:
    """
    Fickle mutation for an experiment. The evolution process in experiment is split in 3 steps:
    On the first step with the help of MAB the `fake_add_mutation` must be chosen,
    since it gives the highest reward, then `fake_add_mutation2` and on the last step -- `fake_add_mutation3`.

    On the first step of evolution this mutation gives reward=3, then reward=-1 and finally reward=1
    """

    if graph.depth >= requirements.max_depth:
        # add mutation is not possible
        return graph

    if len(graph.nodes) < 80:
        nodes_to_add_num = 3
    elif 80 < len(graph.nodes) < 140:
        nodes_to_add_num = -1
    else:
        nodes_to_add_num = 1

    return _apply_mutation(graph=graph, nodes_to_add_num=nodes_to_add_num, requirements=requirements,
                           graph_gen_params=graph_gen_params, parameters=parameters)


def fake_add_mutation2(graph: OptGraph,
                       requirements: GraphRequirements,
                       graph_gen_params: GraphGenerationParams,
                       parameters: AlgorithmParameters) -> OptGraph:
    """
    Fickle mutation for an experiment. The evolution process in experiment is split in 3 steps:
    On the first step with the help of MAB the `fake_add_mutation` must be chosen,
    since it gives the highest reward, then `fake_add_mutation2` and on the last step -- `fake_add_mutation3`.

    On the first step of evolution this mutation gives reward=-1, then reward=2 and finally reward=-1
    """

    if graph.depth >= requirements.max_depth:
        # add mutation is not possible
        return graph

    if len(graph.nodes) < 80:
        nodes_to_add_num = -1
    elif 80 < len(graph.nodes) < 140:
        nodes_to_add_num = 2
    else:
        nodes_to_add_num = -1

    return _apply_mutation(graph=graph, nodes_to_add_num=nodes_to_add_num, requirements=requirements,
                           graph_gen_params=graph_gen_params, parameters=parameters)


def fake_add_mutation3(graph: OptGraph,
                       requirements: GraphRequirements,
                       graph_gen_params: GraphGenerationParams,
                       parameters: AlgorithmParameters) -> OptGraph:
    """
    Fickle mutation for an experiment. The evolution process in experiment is split in 3 steps:
    On the first step with the help of MAB the `fake_add_mutation` must be chosen,
    since it gives the highest reward, then `fake_add_mutation2` and on the last step -- `fake_add_mutation3`.

    On the first step of evolution this mutation gives reward=1, then reward=-3 and finally reward=2
    """

    if graph.depth >= requirements.max_depth:
        # add mutation is not possible
        return graph

    if len(graph.nodes) < 80:
        nodes_to_add_num = 1
    elif 80 < len(graph.nodes) < 140:
        nodes_to_add_num = -3
    else:
        nodes_to_add_num = 2

    return _apply_mutation(graph=graph, nodes_to_add_num=nodes_to_add_num, requirements=requirements,
                           graph_gen_params=graph_gen_params, parameters=parameters)


def _apply_mutation(graph: OptGraph, nodes_to_add_num: int, requirements: GraphRequirements,
                    graph_gen_params: GraphGenerationParams, parameters: AlgorithmParameters):
    """ Adds or deletes nodes according to specified number. """
    if nodes_to_add_num < 0:
        for _ in range(nodes_to_add_num):
            graph = single_drop_mutation(graph, requirements, graph_gen_params, parameters)
    else:
        for _ in range(nodes_to_add_num):
            node_to_mutate = choice(graph.nodes)

            single_add_strategies = [add_as_child, add_separate_parent_node]
            if node_to_mutate.nodes_from:
                single_add_strategies.append(add_intermediate_node)
            strategy = choice(single_add_strategies)

            graph = strategy(graph, node_to_mutate, graph_gen_params.node_factory)

    return graph



from typing import Union, Optional, Sequence

from golem.core.dag.graph import Graph
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.mutation import Mutation
from golem.core.optimisers.genetic.operators.operator import EvaluationOperator, PopulationT
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.core.optimisers.populational_optimizer import PopulationalOptimizer
from golem.core.optimisers.random.random_search import RandomSearchOptimizer
from golem.utilities.data_structures import ensure_wrapped_in_sequence


class PopulationalRandomMutationOptimizer(PopulationalOptimizer):
    """
    Populational random search-based graph models optimizer.
    Implemented as a baseline to compare with evolutionary algorithm.
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Union[Graph, Sequence[Graph]],
                 requirements: Optional[GraphRequirements] = None,
                 graph_generation_params: Optional[GraphGenerationParams] = None,
                 graph_optimizer_params: Optional[GPAlgorithmParameters] = None):
        requirements = requirements or GraphRequirements()
        graph_optimizer_params = graph_optimizer_params or GPAlgorithmParameters()
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        self.initial_individuals = [Individual(graph, metadata=requirements.static_individual_metadata)
                                    for graph in self.initial_graphs]
        self.mutation = Mutation(self.graph_optimizer_params, self.requirements, self.graph_generation_params)

    def _evolve_population(self, evaluator: EvaluationOperator) -> PopulationT:
        new_population = ensure_wrapped_in_sequence(self.mutation(self.population))
        new_population = evaluator(new_population)
        return new_population

    def _initial_population(self, evaluator: EvaluationOperator):
        self._update_population(evaluator(self.initial_individuals), 'initial_assumptions')
        pop_size = self.graph_optimizer_params.pop_size

        if len(self.initial_individuals) < pop_size:
            self.initial_individuals = self._extend_population(self.initial_individuals, pop_size)
            # Adding of extended population to history
            self._update_population(evaluator(self.initial_individuals), 'extended_initial_assumptions')


class RandomMutationOptimizer(RandomSearchOptimizer):
    """
    Random search-based graph models optimizer
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Union[Graph, Sequence[Graph]],
                 requirements: Optional[GraphRequirements] = None,
                 graph_generation_params: Optional[GraphGenerationParams] = None,
                 graph_optimizer_params: Optional[GPAlgorithmParameters] = None):
        requirements = requirements or GraphRequirements()
        graph_optimizer_params = graph_optimizer_params or GPAlgorithmParameters()
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        self.mutation = Mutation(self.graph_optimizer_params, self.requirements, self.graph_generation_params)

    def _generate_new_individual(self) -> Individual:
        return self.mutation(self.best_individual)



import os.path
import _pickle as pickle
import random
import re
from typing import Union, Sequence, Optional

from mabwiser.mab import MAB, LearningPolicy
from scipy.special import softmax

from golem.core.dag.graph import Graph
from golem.core.dag.graph_node import GraphNode
from golem.core.optimisers.adaptive.operator_agent import OperatorAgent, ActType, ObsType, ExperienceBuffer
from golem.core.optimisers.adaptive.reward_agent import FitnessRateRankRewardTransformer
from golem.core.paths import default_data_dir


class MultiArmedBanditAgent(OperatorAgent):
    def __init__(self,
                 actions: Sequence[ActType],
                 n_jobs: int = 1,
                 enable_logging: bool = True,
                 decaying_factor: float = 1.0,
                 path_to_save: Optional[str] = None,
                 is_initial_fit: bool = True):
        super().__init__(actions=actions, enable_logging=enable_logging)
        self.actions = list(actions)
        self._indices = list(range(len(actions)))
        self._arm_by_action = dict(zip(actions, self._indices))
        self._agent = MAB(arms=self._indices,
                          learning_policy=LearningPolicy.EpsilonGreedy(epsilon=0.4),
                          n_jobs=n_jobs)
        self._reward_agent = FitnessRateRankRewardTransformer(decaying_factor=decaying_factor)
        if is_initial_fit:
            self._initial_fit()
        self._path_to_save = path_to_save

    def _initial_fit(self):
        n = len(self.actions)
        uniform_rewards = [1. / n] * n
        self._agent.fit(decisions=self._indices, rewards=uniform_rewards)

    def choose_action(self, obs: ObsType) -> ActType:
        arm = self._agent.predict()
        action = self.actions[arm]
        return action

    def get_action_values(self, obs: Optional[ObsType] = None) -> Sequence[float]:
        prob_dict = self._agent.predict_expectations()
        prob_list = [prob_dict[i] for i in range(len(prob_dict))]
        return prob_list

    def get_action_probs(self, obs: Optional[ObsType] = None) -> Sequence[float]:
        return softmax(self.get_action_values())

    def choose_nodes(self, graph: Graph, num_nodes: int = 1) -> Union[GraphNode, Sequence[GraphNode]]:
        subject_nodes = random.sample(graph.nodes, k=num_nodes)
        return subject_nodes[0] if num_nodes == 1 else subject_nodes

    def partial_fit(self, experience: ExperienceBuffer):
        """Continues learning of underlying agent with new experience."""
        _, arms, processed_rewards = self._get_experience(experience)
        self._agent.partial_fit(decisions=arms, rewards=processed_rewards)

    def _get_experience(self, experience: ExperienceBuffer):
        """ Get experience from ExperienceBuffer, process rewards and log. """
        obs, actions, rewards = experience.retrieve_experience()
        arms = [self._arm_by_action[action] for action in actions]
        processed_rewards = self._reward_agent.get_rewards_for_arms(rewards, arms)
        self._dbg_log(obs, actions, processed_rewards)
        return obs, arms, processed_rewards

    def save(self, path_to_save: Optional[str] = None):
        """ Saves bandit to specified file. """

        if not path_to_save:
            path_to_save = self._path_to_save

        # if path was not specified at all
        if not path_to_save:
            path_to_save = os.path.join(default_data_dir(), 'MAB')

        if not path_to_save.endswith('.pkl'):
            os.makedirs(path_to_save, exist_ok=True)
            mabs_num = [int(name.split('_')[0]) for name in os.listdir(path_to_save)
                        if re.fullmatch(r'\d_mab.pkl', name)]
            if not mabs_num:
                max_saved_mab = 0
            else:
                max_saved_mab = max(mabs_num) + 1
            path_to_file = os.path.join(path_to_save, f'{max_saved_mab}_mab.pkl')
        else:
            path_to_dir = os.path.dirname(path_to_save)
            os.makedirs(path_to_dir, exist_ok=True)
            path_to_file = path_to_save
        with open(path_to_file, 'wb') as f:
            pickle.dump(self, f)
        self._log.info(f"MAB was saved to {path_to_file}")

    @staticmethod
    def load(path: str):
        """ Loads bandit from the specified file. """
        with open(path, 'rb') as f:
            mab = pickle.load(f)
        return mab



from __future__ import annotations

import datetime
import os
from copy import deepcopy
from pathlib import Path
from textwrap import wrap
from typing import Any, Callable, Dict, Iterable, Literal, Optional, Sequence, TYPE_CHECKING, Tuple, Union
from uuid import uuid4

import networkx as nx
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import to_hex
from matplotlib.patches import ArrowStyle
from pyvis.network import Network
from seaborn import color_palette

from golem.core.dag.convert import graph_structure_as_nx_graph
from golem.core.dag.graph_utils import distance_to_primary_level
from golem.core.log import default_log
from golem.core.paths import default_data_dir

if TYPE_CHECKING:
    from golem.core.dag.graph import Graph
    from golem.core.optimisers.graph import OptGraph
    from golem.core.dag.graph_node import GraphNode

    GraphType = Union[Graph, OptGraph]
    GraphConvertType = Callable[[GraphType], Tuple[nx.DiGraph, Dict[uuid4, GraphNode]]]

PathType = Union[os.PathLike, str]

MatplotlibColorType = Union[str, Sequence[float]]
LabelsColorMapType = Dict[str, MatplotlibColorType]
NodeColorFunctionType = Callable[[Iterable[str]], LabelsColorMapType]
NodeColorType = Union[MatplotlibColorType, LabelsColorMapType, NodeColorFunctionType]


class GraphVisualizer:
    def __init__(self, graph: GraphType, visuals_params: Optional[Dict[str, Any]] = None,
                 to_nx_convert_func: GraphConvertType = graph_structure_as_nx_graph):
        visuals_params = visuals_params or {}
        default_visuals_params = dict(
            engine='matplotlib',
            dpi=100,
            node_color=self._get_colors_by_labels,
            node_size_scale=1.0,
            font_size_scale=1.0,
            edge_curvature_scale=1.0,
            node_names_placement='auto',
            nodes_layout_function=GraphVisualizer._get_hierarchy_pos_by_distance_to_primary_level,
            figure_size=(7, 7),
            save_path=None,
        )
        default_visuals_params.update(visuals_params)
        self.visuals_params = default_visuals_params
        self.to_nx_convert_func = to_nx_convert_func
        self._update_graph(graph)
        self.log = default_log(self)

    def _update_graph(self, graph: GraphType):
        self.graph = graph
        self.nx_graph, self.nodes_dict = self.to_nx_convert_func(self.graph)

    def visualise(self, save_path: Optional[PathType] = None, engine: Optional[str] = None,
                  node_color: Optional[NodeColorType] = None, dpi: Optional[int] = None,
                  node_size_scale: Optional[float] = None, font_size_scale: Optional[float] = None,
                  edge_curvature_scale: Optional[float] = None, figure_size: Optional[Tuple[int, int]] = None,
                  nodes_labels: Dict[int, str] = None, edges_labels: Dict[int, str] = None,
                  node_names_placement: Optional[Literal['auto', 'nodes', 'legend', 'none']] = None,
                  nodes_layout_function: Optional[Callable[[nx.DiGraph], Dict[Any, Tuple[float, float]]]] = None,
                  title: Optional[str] = None):
        engine = engine or self._get_predefined_value('engine')

        if not self.graph.nodes:
            raise ValueError('Empty graph can not be visualized.')

        if engine == 'matplotlib':
            self._draw_with_networkx(save_path=save_path, node_color=node_color, dpi=dpi,
                                     node_size_scale=node_size_scale, font_size_scale=font_size_scale,
                                     edge_curvature_scale=edge_curvature_scale, figure_size=figure_size,
                                     title=title, nodes_labels=nodes_labels, edges_labels=edges_labels,
                                     nodes_layout_function=nodes_layout_function,
                                     node_names_placement=node_names_placement)
        elif engine == 'pyvis':
            self._draw_with_pyvis(save_path, node_color)
        elif engine == 'graphviz':
            self._draw_with_graphviz(save_path, node_color, dpi)
        else:
            raise NotImplementedError(f'Unexpected visualization engine: {engine}. '
                                      'Possible values: matplotlib, pyvis, graphviz.')

    def draw_nx_dag(
            self, ax: Optional[plt.Axes] = None, node_color: Optional[NodeColorType] = None,
            node_size_scale: Optional[float] = None, font_size_scale: Optional[float] = None,
            edge_curvature_scale: Optional[float] = None, nodes_labels: Dict[int, str] = None,
            edges_labels: Dict[int, str] = None,
            nodes_layout_function: Optional[Callable[[nx.DiGraph], Dict[Any, Tuple[float, float]]]] = None,
            node_names_placement: Optional[Literal['auto', 'nodes', 'legend', 'none']] = None):
        node_color = node_color or self._get_predefined_value('node_color')
        node_size_scale = node_size_scale or self._get_predefined_value('node_size_scale')
        font_size_scale = font_size_scale or self._get_predefined_value('font_size_scale')
        edge_curvature_scale = (edge_curvature_scale if edge_curvature_scale is not None
                                else self._get_predefined_value('edge_curvature_scale'))
        nodes_layout_function = nodes_layout_function or self._get_predefined_value('nodes_layout_function')
        node_names_placement = node_names_placement or self._get_predefined_value('node_names_placement')

        nx_graph, nodes = self.nx_graph, self.nodes_dict

        if ax is None:
            ax = plt.gca()

        # Define colors
        if callable(node_color):
            node_color = node_color([str(node) for node in nodes.values()])
        if isinstance(node_color, dict):
            node_color = [node_color.get(str(node), node_color.get(None)) for node in nodes.values()]
        else:
            node_color = [node_color for _ in nodes]
        # Get node positions
        if nodes_layout_function == GraphVisualizer._get_hierarchy_pos_by_distance_to_primary_level:
            pos = nodes_layout_function(nx_graph, nodes)
        else:
            pos = nodes_layout_function(nx_graph)

        node_size = self._get_scaled_node_size(len(nodes), node_size_scale)

        with_node_names = node_names_placement != 'none'

        if node_names_placement in ('auto', 'none'):
            node_names_placement = GraphVisualizer._define_node_names_placement(node_size)

        if node_names_placement == 'nodes':
            self._draw_nx_big_nodes(ax, pos, nodes, node_color, node_size, font_size_scale, with_node_names)
        elif node_names_placement == 'legend':
            self._draw_nx_small_nodes(ax, pos, nodes, node_color, node_size, font_size_scale, with_node_names)
        self._draw_nx_curved_edges(ax, pos, node_size, edge_curvature_scale)
        self._draw_nx_labels(ax, pos, font_size_scale, nodes_labels, edges_labels)

    def _get_predefined_value(self, param: str):
        if param not in self.visuals_params:
            self.log.warning(f'No default param found: {param}.')
        return self.visuals_params.get(param)

    def _draw_with_networkx(
            self, save_path: Optional[PathType] = None, node_color: Optional[NodeColorType] = None,
            dpi: Optional[int] = None, node_size_scale: Optional[float] = None,
            font_size_scale: Optional[float] = None, edge_curvature_scale: Optional[float] = None,
            figure_size: Optional[Tuple[int, int]] = None, title: Optional[str] = None,
            nodes_labels: Dict[int, str] = None, edges_labels: Dict[int, str] = None,
            nodes_layout_function: Optional[Callable[[nx.DiGraph], Dict[Any, Tuple[float, float]]]] = None,
            node_names_placement: Optional[Literal['auto', 'nodes', 'legend', 'none']] = None):
        save_path = save_path or self._get_predefined_value('save_path')
        node_color = node_color or self._get_predefined_value('node_color')
        dpi = dpi or self._get_predefined_value('dpi')
        figure_size = figure_size or self._get_predefined_value('figure_size')

        ax = GraphVisualizer._setup_matplotlib_figure(figure_size, dpi, title)
        self.draw_nx_dag(ax, node_color, node_size_scale, font_size_scale, edge_curvature_scale,
                         nodes_labels, edges_labels, nodes_layout_function, node_names_placement)
        GraphVisualizer._rescale_matplotlib_figure(ax)
        if not save_path:
            plt.show()
        else:
            plt.savefig(save_path, dpi=dpi)
            plt.close()

    def _draw_with_pyvis(self, save_path: Optional[PathType] = None, node_color: Optional[NodeColorType] = None):
        save_path = save_path or self._get_predefined_value('save_path')
        node_color = node_color or self._get_predefined_value('node_color')

        net = Network('500px', '1000px', directed=True)
        nx_graph, nodes = self.nx_graph, self.nodes_dict
        node_color = self._define_colors(node_color, nodes)
        for n, data in nx_graph.nodes(data=True):
            operation = nodes[n]
            label = str(operation)
            data['n_id'] = str(n)
            data['label'] = label.replace('_', ' ')
            params = operation.content.get('params')
            if isinstance(params, dict):
                params = str(params)[1:-1]
            data['title'] = params
            data['level'] = distance_to_primary_level(operation)
            data['color'] = to_hex(node_color.get(label, node_color.get(None)))
            data['font'] = '20px'
            data['labelHighlightBold'] = True

        for _, data in nx_graph.nodes(data=True):
            net.add_node(**data)
        for u, v in nx_graph.edges:
            net.add_edge(str(u), str(v))

        if save_path:
            net.save_graph(str(save_path))
            return
        save_path = Path(default_data_dir(), 'graph_plots', str(uuid4()) + '.html')
        save_path.parent.mkdir(exist_ok=True)
        net.show(str(save_path))
        remove_old_files_from_dir(save_path.parent)

    def _draw_with_graphviz(self, save_path: Optional[PathType] = None, node_color: Optional[NodeColorType] = None,
                            dpi: Optional[int] = None):
        save_path = save_path or self._get_predefined_value('save_path')
        node_color = node_color or self._get_predefined_value('node_color')
        dpi = dpi or self._get_predefined_value('dpi')

        nx_graph, nodes = self.nx_graph, self.nodes_dict
        node_color = self._define_colors(node_color, nodes)
        for n, data in nx_graph.nodes(data=True):
            label = str(nodes[n])
            data['label'] = label.replace('_', ' ')
            data['color'] = to_hex(node_color.get(label, node_color.get(None)))

        gv_graph = nx.nx_agraph.to_agraph(nx_graph)
        kwargs = {'prog': 'dot', 'args': f'-Gnodesep=0.5 -Gdpi={dpi} -Grankdir="LR"'}

        if save_path:
            gv_graph.draw(save_path, **kwargs)
        else:
            save_path = Path(default_data_dir(), 'graph_plots', str(uuid4()) + '.png')
            save_path.parent.mkdir(exist_ok=True)
            gv_graph.draw(save_path, **kwargs)

            img = plt.imread(str(save_path))
            plt.imshow(img)
            plt.gca().axis('off')
            plt.gcf().set_dpi(dpi)
            plt.tight_layout()
            plt.show()
            remove_old_files_from_dir(save_path.parent)

    @staticmethod
    def _get_scaled_node_size(nodes_amount: int, size_scale: float) -> float:
        min_size = 150
        max_size = 12000
        size = max(max_size * (1 - np.log10(nodes_amount)), min_size)
        return size * size_scale

    @staticmethod
    def _get_scaled_font_size(nodes_amount: int, size_scale: float) -> float:
        min_size = 14
        max_size = 30
        size = max(max_size * (1 - np.log10(nodes_amount)), min_size)
        return size * size_scale

    @staticmethod
    def _get_colors_by_labels(labels: Iterable[str]) -> LabelsColorMapType:
        unique_labels = list(set(labels))
        palette = color_palette('tab10', len(unique_labels))
        return {label: palette[unique_labels.index(label)] for label in labels}

    @staticmethod
    def _define_colors(node_color, nodes):
        if callable(node_color):
            colors = node_color([str(node) for node in nodes.values()])
        elif isinstance(node_color, dict):
            colors = node_color
        else:
            colors = {str(node): node_color for node in nodes.values()}
        return colors

    @staticmethod
    def _setup_matplotlib_figure(figure_size: Tuple[float, float], dpi: int, title: Optional[str] = None) -> plt.Axes:
        fig, ax = plt.subplots(figsize=figure_size)
        fig.set_dpi(dpi)
        plt.title(title)
        return ax

    @staticmethod
    def _rescale_matplotlib_figure(ax):
        """Rescale the figure for all nodes to fit in."""

        x_1, x_2 = ax.get_xlim()
        y_1, y_2 = ax.get_ylim()
        offset = 0.2
        x_offset = x_2 * offset
        y_offset = y_2 * offset
        ax.set_xlim(x_1 - x_offset, x_2 + x_offset)
        ax.set_ylim(y_1 - y_offset, y_2 + y_offset)
        ax.axis('off')
        plt.tight_layout()

    def _draw_nx_big_nodes(self, ax, pos, nodes, node_color, node_size, font_size_scale, with_node_names):
        # Draw the graph's nodes.
        nx.draw_networkx_nodes(self.nx_graph, pos, node_size=node_size, ax=ax, node_color='w', linewidths=3,
                               edgecolors=node_color)
        if not with_node_names:
            return
        # Draw the graph's node labels.
        node_labels = {node_id: str(node) for node_id, node in nodes.items()}
        font_size = GraphVisualizer._get_scaled_font_size(len(nodes), font_size_scale)
        for node, (x, y) in pos.items():
            text = '\n'.join(wrap(node_labels[node].replace('_', ' ').replace('-', ' '), 10))
            ax.text(x, y, text,
                    ha='center', va='center',
                    fontsize=font_size,
                    bbox=dict(alpha=0.9, color='w', boxstyle='round'))

    def _draw_nx_small_nodes(self, ax, pos, nodes, node_color, node_size, font_size_scale, with_node_names):
        nx_graph = self.nx_graph
        markers = 'os^>v<dph8'
        label_markers = {}
        labels_added = set()
        color_counts = {}
        for node_num, (node_id, node) in enumerate(nodes.items()):
            label = str(node)
            color = node_color[node_num]
            color = to_hex(color, keep_alpha=True)  # Convert the color to a hashable type.
            marker = label_markers.get(label)
            if marker is None:
                color_count = color_counts.get(color, 0)
                if color_count > len(markers) - 1:
                    self.log.warning(f'Too much node labels derive the same color: {color}. The markers may repeat.\n'
                                     '\tSpecify the parameter "node_color" to set distinct colors.')
                    color_count = color_count % len(markers)
                marker = markers[color_count]
                label_markers[label] = marker
                color_counts[color] = color_count + 1
            nx.draw_networkx_nodes(nx_graph, pos, [node_id], ax=ax, node_color=[color], node_size=node_size,
                                   node_shape=marker)
            if label in labels_added:
                continue
            ax.plot([], [], marker=marker, linestyle='None', color=color, label=label)
            labels_added.add(label)
        if not with_node_names:
            return
        # @morrisnein took the following code from https://stackoverflow.com/a/27512450
        handles, labels = ax.get_legend_handles_labels()
        # Sort both labels and handles by labels
        labels, handles = zip(*sorted(zip(labels, handles), key=lambda t: t[0]))
        ax.legend(handles, labels, prop={'size': round(20 * font_size_scale)})

    def _draw_nx_curved_edges(self, ax, pos, node_size, edge_curvature_scale):
        nx_graph = self.nx_graph
        # The ongoing section defines curvature for all edges.
        #   This is 'connection style' for an edge that does not intersect any nodes.
        connection_style = 'arc3'
        #   This is 'connection style' template for an edge that is too close to any node and must bend around it.
        #   The curvature value is defined individually for each edge.
        connection_style_curved_template = connection_style + ',rad={}'
        default_edge_curvature = 0.3
        #   The minimum distance from a node to an edge on which the edge must bend around the node.
        node_distance_gap = 0.15
        for u, v, e in nx_graph.edges(data=True):
            e['connectionstyle'] = connection_style
            p_1, p_2 = np.array(pos[u]), np.array(pos[v])
            p_1_2 = p_2 - p_1
            p_1_2_length = np.linalg.norm(p_1_2)
            # Finding the closest node to the edge.
            min_distance_found = node_distance_gap * 2  # It just must be bigger than the gap.
            closest_node_id = None
            for node_id in nx_graph.nodes:
                if node_id in (u, v):
                    continue  # The node is adjacent to the edge.
                p_3 = np.array(pos[node_id])
                distance_to_node = abs(np.cross(p_1_2, p_3 - p_1)) / p_1_2_length
                if (distance_to_node > min(node_distance_gap, min_distance_found)  # The node is too far.
                        or ((p_3 - p_1) @ p_1_2) < 0  # There's no perpendicular from the node to the edge.
                        or ((p_3 - p_2) @ -p_1_2) < 0):
                    continue
                min_distance_found = distance_to_node
                closest_node_id = node_id

            if closest_node_id is None:
                continue  # There's no node to bend around.
            # Finally, define the edge's curvature based on the closest node position.
            p_3 = np.array(pos[closest_node_id])
            p_1_3 = p_3 - p_1
            curvature_strength = default_edge_curvature * edge_curvature_scale
            # 'alpha' denotes the angle between the abscissa and the edge.
            cos_alpha = p_1_2[0] / p_1_2_length
            sin_alpha = np.sqrt(1 - cos_alpha ** 2) * (-1) ** (p_1_2[1] < 0)
            # The closest node is placed as if the edge matched the abscissa.
            # Then, its ordinate shows on which side of the edge it is, "on the left" or "on the right".
            rotation_matrix = np.array([[cos_alpha, sin_alpha], [-sin_alpha, cos_alpha]])
            p_1_3_rotated = rotation_matrix @ p_1_3
            curvature_direction = (-1) ** (p_1_3_rotated[1] < 0)  # +1 is a "cup" \/, -1 is a "cat" /\.
            edge_curvature = curvature_direction * curvature_strength
            e['connectionstyle'] = connection_style_curved_template.format(edge_curvature)
            # Define edge center position for labels.
            edge_center_position = np.mean([p_1, p_2], axis=0)
            edge_curvature_shift = np.linalg.inv(rotation_matrix) @ [0, -1 * edge_curvature / 4]
            edge_center_position += edge_curvature_shift
            e['edge_center_position'] = edge_center_position
        # Draw the graph's edges.
        arrow_style = ArrowStyle('Simple', head_length=1.5, head_width=0.8)
        for u, v, e in nx_graph.edges(data=True):
            nx.draw_networkx_edges(nx_graph, pos, edgelist=[(u, v)], node_size=node_size, ax=ax, arrowsize=10,
                                   arrowstyle=arrow_style, connectionstyle=e['connectionstyle'])
        self._rescale_matplotlib_figure(ax)

    def _draw_nx_labels(self, ax: plt.Axes, pos: Any, font_size_scale: float,
                        nodes_labels: Dict[int, str], edges_labels: Dict[int, str]):
        """ Set labels with scores to nodes and edges. """

        def calculate_labels_bias(ax: plt.Axes, y_span: int):
            y_1, y_2 = ax.get_ylim()
            y_size = y_2 - y_1
            if y_span == 1:
                bias_scale = 0.25  # Fits between the central line and the upper bound.
            else:
                bias_scale = 1 / y_span / 3 * 0.5  # Fits between the narrowest horizontal rows.
            bias = y_size * bias_scale
            return bias

        def match_labels_with_nx_nodes(nx_graph: nx.DiGraph, labels: Dict[int, str]) -> Dict[str, str]:
            """ Matches index of node in GOLEM graph with networkx node name. """
            nx_nodes = list(nx_graph.nodes.keys())
            nx_labels = {}
            for index, label in labels.items():
                nx_labels[nx_nodes[index]] = label
            return nx_labels

        def match_labels_with_nx_edges(nx_graph: nx.DiGraph, labels: Dict[int, str]) \
                -> Dict[Tuple[str, str], str]:
            """ Matches index of edge in GOLEM graph with tuple of networkx nodes names. """
            nx_nodes = list(nx_graph.nodes.keys())
            edges = self.graph.get_edges()
            nx_labels = {}
            for index, label in labels.items():
                edge = edges[index]
                parent_node_nx = nx_nodes[self.graph.nodes.index(edge[0])]
                child_node_nx = nx_nodes[self.graph.nodes.index(edge[1])]
                nx_labels[(parent_node_nx, child_node_nx)] = label
            return nx_labels

        def draw_node_labels(node_labels, ax, bias, font_size, nx_graph, pos):
            labels_pos = deepcopy(pos)
            for value in labels_pos.values():
                value[1] += bias
            bbox = dict(alpha=0.9, color='w')

            nodes_nx_labels = match_labels_with_nx_nodes(nx_graph=nx_graph, labels=node_labels)
            nx.draw_networkx_labels(
                nx_graph, labels_pos,
                ax=ax,
                labels=nodes_nx_labels,
                font_color='black',
                font_size=font_size,
                bbox=bbox
            )

        def draw_edge_labels(edge_labels, ax, bias, font_size, nx_graph, pos):
            labels_pos_edges = deepcopy(pos)
            label_bias_y = 2 / 3 * bias
            if len(set([coord[1] for coord in pos.values()])) == 1 and len(list(pos.values())) > 2:
                for value in labels_pos_edges.values():
                    value[1] += label_bias_y
            edges_nx_labels = match_labels_with_nx_edges(nx_graph=nx_graph, labels=edge_labels)
            bbox = dict(alpha=0.9, color='w')
            # Set labels for edges
            for u, v, e in nx_graph.edges(data=True):
                if (u, v) not in edges_nx_labels:
                    continue
                current_pos = labels_pos_edges
                if 'edge_center_position' in e:
                    x, y = e['edge_center_position']
                    plt.text(x, y, edges_nx_labels[(u, v)], bbox=bbox, fontsize=font_size)
                else:
                    nx.draw_networkx_edge_labels(
                        nx_graph, current_pos, {(u, v): edges_nx_labels[(u, v)]},
                        label_pos=0.5, ax=ax,
                        font_color='black',
                        font_size=font_size,
                        rotate=False,
                        bbox=bbox
                    )

        if not (edges_labels or nodes_labels):
            return

        nodes_amount = len(pos)
        font_size = GraphVisualizer._get_scaled_font_size(nodes_amount, font_size_scale * 0.75)
        _, y_span = GraphVisualizer._get_x_y_span(pos)
        bias = calculate_labels_bias(ax, y_span)

        if nodes_labels:
            draw_node_labels(nodes_labels, ax, bias, font_size, self.nx_graph, pos)

        if edges_labels:
            draw_edge_labels(edges_labels, ax, bias, font_size, self.nx_graph, pos)

    @staticmethod
    def _get_hierarchy_pos_by_distance_to_primary_level(nx_graph: nx.DiGraph, nodes: Dict
                                                        ) -> Dict[Any, Tuple[float, float]]:
        """By default, returns 'networkx.multipartite_layout' positions based on 'hierarchy_level`
        from node data - the property must be set beforehand.
        :param graph: the graph.
        """
        for node_id, node_data in nx_graph.nodes(data=True):
            node_data['hierarchy_level'] = distance_to_primary_level(nodes[node_id])

        return nx.multipartite_layout(nx_graph, subset_key='hierarchy_level')

    @staticmethod
    def _get_x_y_span(pos: Dict[Any, Tuple[float, float]]) -> Tuple[int, int]:
        pos_x, pos_y = np.split(np.array(tuple(pos.values())), 2, axis=1)
        x_span = max(pos_x) - min(pos_x)
        y_span = max(pos_y) - min(pos_y)
        return x_span, y_span

    @staticmethod
    def _define_node_names_placement(node_size):
        if node_size >= 1000:
            node_names_placement = 'nodes'
        else:
            node_names_placement = 'legend'
        return node_names_placement


def remove_old_files_from_dir(dir_: Path, time_interval=datetime.timedelta(minutes=10)):
    for path in dir_.iterdir():
        if datetime.datetime.now() - datetime.datetime.fromtimestamp(path.stat().st_ctime) > time_interval:
            path.unlink()



from abc import ABC, abstractmethod
from enum import Enum
from os import PathLike
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple, TypeVar, Union

import networkx as nx

from golem.core.dag.graph_node import GraphNode
from golem.visualisation.graph_viz import GraphVisualizer, NodeColorType

NodeType = TypeVar('NodeType', bound=GraphNode, covariant=False, contravariant=False)


class ReconnectType(Enum):
    """Defines allowed kinds of removals in Graph. Used by mutations."""
    none = 'none'  # do not reconnect predecessors
    single = 'single'  # reconnect a predecessor only if it's single
    all = 'all'  # reconnect all predecessors to all successors


class Graph(ABC):
    """Defines abstract graph interface that's required by graph optimisation process.
    """

    @abstractmethod
    def add_node(self, node: GraphNode):
        """Adds new node to the graph together with its parent nodes.

        Args:
            node: graph nodes
        """
        raise NotImplementedError()

    @abstractmethod
    def update_node(self, old_node: GraphNode, new_node: GraphNode):
        """Replaces ``old_node`` node with ``new_node``

        Args:
            old_node: node to be replaced
            new_node: node to be placed instead
        """
        raise NotImplementedError()

    @abstractmethod
    def update_subtree(self, old_subtree: GraphNode, new_subtree: GraphNode):
        """Changes ``old_subtree`` subtree to ``new_subtree``

        Args:
            old_subtree: node and its subtree to be removed
            new_subtree: node and its subtree to be placed instead
        """
        raise NotImplementedError()

    @abstractmethod
    def delete_node(self, node: GraphNode, reconnect: ReconnectType = ReconnectType.single):
        """Removes ``node`` from the graph.
        If ``node`` has only one child, then connects all of the ``node`` parents to it.

        Args:
            node: node of the graph to be deleted
            reconnect: defines how to treat left edges between parents and children
        """
        raise NotImplementedError()

    @abstractmethod
    def delete_subtree(self, subtree: GraphNode):
        """Deletes given node with all its parents.
        Deletes all edges from removed nodes to remaining graph nodes

        Args:
            subtree: node to be deleted with all of its parents
                and their connections amongst the remaining graph nodes
        """
        raise NotImplementedError()

    @abstractmethod
    def node_children(self, node: GraphNode) -> Sequence[Optional[GraphNode]]:
        """Returns all children of the ``node``

        Args:
            node: for getting children from

        Returns: children of the ``node``
        """
        raise NotImplementedError()

    @abstractmethod
    def connect_nodes(self, node_parent: GraphNode, node_child: GraphNode):
        """Adds edge between ``parent`` and ``child``

        Args:
            node_parent: acts like parent in graph connection relations
            node_child:  acts like child in graph connection relations
        """
        raise NotImplementedError()

    @abstractmethod
    def disconnect_nodes(self, node_parent: GraphNode, node_child: GraphNode,
                         clean_up_leftovers: bool = False):
        """Removes an edge between two nodes

        Args:
            node_parent: where the removing edge comes out
            node_child: where the removing edge enters
            clean_up_leftovers: whether to remove the remaining invalid vertices with edges or not
        """
        raise NotImplementedError()

    @abstractmethod
    def get_edges(self) -> Sequence[Tuple[GraphNode, GraphNode]]:
        """Gets all available edges in this graph

        Returns:
            pairs of parent_node -> child_node
        """
        raise NotImplementedError()

    def get_nodes_by_name(self, name: str) -> List[GraphNode]:
        """Returns list of nodes with the required ``name``

        Args:
            name: name to filter by

        Returns:
            list: relevant nodes (empty if there are no such nodes)
        """

        appropriate_nodes = filter(lambda x: x.name == name, self.nodes)

        return list(appropriate_nodes)

    def get_node_by_uid(self, uid: str) -> Optional[GraphNode]:
        """Returns node with the required ``uid``

        Args:
            uid: uid of node to filter by

        Returns:
            Optional[Node]: relevant node (None if there is no such node)
        """

        appropriate_nodes = list(filter(lambda x: x.uid == uid, self.nodes))

        return appropriate_nodes[0] if appropriate_nodes else None

    @abstractmethod
    def __eq__(self, other_graph: 'Graph') -> bool:
        """Compares this graph with the ``other_graph``

        Args:
            other_graph: another graph

        Returns:
            is it equal to ``other_graph`` in terms of the graphs
        """
        raise NotImplementedError()

    def root_nodes(self) -> Sequence[GraphNode]:
        raise NotImplementedError()

    @property
    def root_node(self) -> Union[GraphNode, Sequence[GraphNode]]:
        """Gets the final layer node(s) of the graph

        Returns:
            the final layer node(s)
        """
        roots = self.root_nodes()
        if len(roots) == 1:
            return roots[0]
        return roots

    @property
    @abstractmethod
    def nodes(self) -> List[GraphNode]:
        """Return list of all graph nodes

        Returns:
            graph nodes
        """
        raise NotImplementedError()

    @nodes.setter
    @abstractmethod
    def nodes(self, new_nodes: List[GraphNode]):
        raise NotImplementedError()

    @property
    @abstractmethod
    def depth(self) -> int:
        """Gets this graph depth from its sink-node to its source-node

        Returns:
            length of a path from the root node to the farthest primary node
        """
        raise NotImplementedError()

    @property
    def length(self) -> int:
        """Return size of the graph (number of nodes)

        Returns:
            graph size
        """

        return len(self.nodes)

    def show(self, save_path: Optional[Union[PathLike, str]] = None, engine: Optional[str] = None,
             node_color: Optional[NodeColorType] = None, dpi: Optional[int] = None,
             node_size_scale: Optional[float] = None, font_size_scale: Optional[float] = None,
             edge_curvature_scale: Optional[float] = None,
             title: Optional[str] = None,
             node_names_placement: Optional[Literal['auto', 'nodes', 'legend', 'none']] = None,
             nodes_labels: Dict[int, str] = None, edges_labels: Dict[int, str] = None,
             nodes_layout_function: Optional[Callable[[nx.DiGraph], Dict[Any, Tuple[float, float]]]] = None):
        """Visualizes graph or saves its picture to the specified ``path``

        Args:
            save_path: optional, save location of the graph visualization image.
            engine: engine to visualize the graph. Possible values: 'matplotlib', 'pyvis', 'graphviz'.
            node_color: color of nodes to use.
            node_size_scale: use to make node size bigger or lesser. Supported only for the engine 'matplotlib'.
            font_size_scale: use to make font size bigger or lesser. Supported only for the engine 'matplotlib'.
            edge_curvature_scale: use to make edges more or less curved. Supported only for the engine 'matplotlib'.
            dpi: DPI of the output image. Not supported for the engine 'pyvis'.
            title: title for plot
            node_names_placement: variant of node names displaying. Defaults to ``auto``.

                Possible options:

                    - ``auto`` -> empirical rule by node size

                    - ``nodes`` -> place node names on top of the nodes

                    - ``legend`` -> place node names at the legend

                    - ``none`` -> do not show node names

            nodes_labels: labels to display near nodes
            edges_labels: labels to display near edges
            nodes_layout_function: any of `Networkx layout functions \
                <https://networkx.org/documentation/stable/reference/drawing.html#module-networkx.drawing.layout>`_ .
        """
        GraphVisualizer(graph=self) \
            .visualise(save_path=save_path, engine=engine, node_color=node_color, dpi=dpi,
                       node_size_scale=node_size_scale, font_size_scale=font_size_scale,
                       edge_curvature_scale=edge_curvature_scale, node_names_placement=node_names_placement,
                       title=title, nodes_layout_function=nodes_layout_function,
                       nodes_labels=nodes_labels, edges_labels=edges_labels)

    @property
    def graph_description(self) -> Dict:
        """Return summary characteristics of the graph

        Returns:
            dict: containing information about the graph
        """
        return {
            'depth': self.depth,
            'length': self.length,
            'nodes': self.nodes,
        }

    @property
    def descriptive_id(self) -> str:
        """Returns human-readable identifier of the graph.

        Returns:
            str: text description of the content in the node and its parameters
        """
        if self.root_nodes:
            return self.root_node.descriptive_id
        else:
            return sorted(self.nodes, key=lambda x: x.uid)[0].descriptive_id

    def __str__(self):
        return str(self.graph_description)

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return self.length



from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union, Callable, Sequence

from networkx import graph_edit_distance, set_node_attributes

from golem.core.dag.convert import graph_structure_as_nx_graph
from golem.core.dag.graph import Graph, ReconnectType
from golem.core.dag.graph_node import GraphNode
from golem.core.dag.graph_utils import ordered_subnodes_hierarchy, node_depth, graph_has_cycle
from golem.core.paths import copy_doc
from golem.utilities.data_structures import ensure_wrapped_in_sequence, Copyable, remove_items

NodePostprocessCallable = Callable[[Graph, Sequence[GraphNode]], Any]


class LinkedGraph(Graph, Copyable):
    """Graph implementation based on linked graph node
    that directly stores its parent nodes.

    Args:
        nodes: nodes of the Graph
        postprocess_nodes: nodes postprocessing function used after their modification
    """

    def __init__(self, nodes: Union[GraphNode, Sequence[GraphNode]] = (),
                 postprocess_nodes: Optional[NodePostprocessCallable] = None):
        self._nodes = []
        for node in ensure_wrapped_in_sequence(nodes):
            self.add_node(node)
        self._postprocess_nodes = postprocess_nodes or self._empty_postprocess

    @staticmethod
    def _empty_postprocess(*args):
        pass

    @copy_doc(Graph.delete_node)
    def delete_node(self, node: GraphNode, reconnect: ReconnectType = ReconnectType.single) -> object:
        node_children_cached = self.node_children(node)

        self._nodes.remove(node)
        for node_child in node_children_cached:
            node_child.nodes_from.remove(node)

        if reconnect == ReconnectType.single:
            # if removed node had a single child
            # then reconnect it to preceding parent nodes.
            if node.nodes_from and len(node_children_cached) == 1:
                child = node_children_cached[0]
                child.nodes_from.extend(node.nodes_from)
        elif reconnect == ReconnectType.all:
            if node.nodes_from:
                for child in node_children_cached:
                    child.nodes_from.extend(node.nodes_from)
        elif reconnect == ReconnectType.none:
            pass

        self._postprocess_nodes(self, self._nodes)

    @copy_doc(Graph.delete_subtree)
    def delete_subtree(self, subtree: GraphNode):
        subtree_nodes = ordered_subnodes_hierarchy(subtree)
        self._nodes = remove_items(self._nodes, subtree_nodes)
        # prune all edges coming from the removed subtree
        for subtree in self._nodes:
            subtree.nodes_from = remove_items(subtree.nodes_from, subtree_nodes)

    @copy_doc(Graph.update_node)
    def update_node(self, old_node: GraphNode, new_node: GraphNode):
        self.actualise_old_node_children(old_node, new_node)
        new_node.nodes_from.extend(old_node.nodes_from)
        self._nodes.remove(old_node)
        self.add_node(new_node)
        self.sort_nodes()
        self._postprocess_nodes(self, self._nodes)

    @copy_doc(Graph.update_subtree)
    def update_subtree(self, old_subtree: GraphNode, new_subtree: GraphNode):
        new_subtree = deepcopy(new_subtree)
        self.actualise_old_node_children(old_subtree, new_subtree)
        self.delete_subtree(old_subtree)
        self.add_node(new_subtree)
        self.sort_nodes()

    @copy_doc(Graph.add_node)
    def add_node(self, node: GraphNode):
        if node not in self._nodes:
            self._nodes.append(node)
            for n in node.nodes_from:
                self.add_node(n)

    def actualise_old_node_children(self, old_node: GraphNode, new_node: GraphNode):
        """Changes parent of ``old_node`` children to ``new_node``

        :param old_node: node to take children from
        :param new_node: new parent of ``old_node`` children
        """
        old_node_offspring = self.node_children(old_node)
        for old_node_child in old_node_offspring:
            updated_index = old_node_child.nodes_from.index(old_node)
            old_node_child.nodes_from[updated_index] = new_node

    def sort_nodes(self):
        """ Layer by layer sorting """
        if not isinstance(self.root_node, Sequence) and not graph_has_cycle(self):
            self._nodes = ordered_subnodes_hierarchy(self.root_node)

    @copy_doc(Graph.node_children)
    def node_children(self, node: GraphNode) -> List[Optional[GraphNode]]:
        """ Returns list of children of specified node. """
        return [other_node for other_node in self._nodes
                if other_node.nodes_from and
                node in other_node.nodes_from]

    @copy_doc(Graph.connect_nodes)
    def connect_nodes(self, node_parent: GraphNode, node_child: GraphNode):
        if node_child in self.node_children(node_parent):
            return
        node_child.nodes_from.append(node_parent)

    def _clean_up_leftovers(self, node: GraphNode):
        """Removes nodes and edges that do not affect the result of the graph.
        Leftovers are edges and nodes that remain after the removal of the edge / node
            and do not affect the result of the graph.

        :param node: node to be deleted with all of its parents
        """

        if not self.node_children(node):
            self._nodes.remove(node)
            for node in node.nodes_from:
                self._clean_up_leftovers(node)

    @copy_doc(Graph.disconnect_nodes)
    def disconnect_nodes(self, node_parent: GraphNode, node_child: GraphNode,
                         clean_up_leftovers: bool = False):
        if node_parent not in node_child.nodes_from:
            return
        if node_parent not in self._nodes or node_child not in self._nodes:
            return
        node_child.nodes_from.remove(node_parent)
        if clean_up_leftovers:
            self._clean_up_leftovers(node_parent)
        self._postprocess_nodes(self, self._nodes)

    def root_nodes(self) -> Sequence[GraphNode]:
        return [node for node in self._nodes if not any(self.node_children(node))]

    @property
    def nodes(self) -> List[GraphNode]:
        return self._nodes

    @nodes.setter
    def nodes(self, new_nodes: List[GraphNode]):
        self._nodes = new_nodes

    @copy_doc(Graph.__eq__)
    def __eq__(self, other_graph: Graph) -> bool:
        return \
            set(rn.descriptive_id for rn in self.root_nodes()) == \
            set(rn.descriptive_id for rn in other_graph.root_nodes())

    @copy_doc(Graph.descriptive_id)
    @property
    def descriptive_id(self) -> str:
        if self.length == 0:
            return 'EMPTY'
        elif self.root_nodes():
            return ''.join([r.descriptive_id for r in self.root_nodes()])
        else:
            return sorted(self.nodes, key=lambda x: x.uid)[0].descriptive_id

    @copy_doc(Graph.depth)
    @property
    def depth(self) -> int:
        if not self._nodes:
            return 0
        elif not self.root_nodes() or graph_has_cycle(self):
            return -1
        else:
            depths = node_depth(self.root_nodes())
            return max(ensure_wrapped_in_sequence(depths))

    @copy_doc(Graph.get_edges)
    def get_edges(self) -> Sequence[Tuple[GraphNode, GraphNode]]:
        edges = []
        for node in self._nodes:
            if node.nodes_from:
                for parent_node in node.nodes_from:
                    edges.append((parent_node, node))
        return edges


def get_distance_between(graph_1: Graph, graph_2: Graph) -> int:
    """
    Gets edit distance from ``graph_1`` graph to the ``graph_2``

    :param graph_1: left object to compare
    :param graph_2: right object to compare

    :return: graph edit distance (aka Levenstein distance for graphs)
    """

    def node_match(node_data_1: Dict[str, GraphNode], node_data_2: Dict[str, GraphNode]) -> bool:
        """Checks if the two given nodes are identical

        :param node_data_1: nx_graph format for the first node to compare
        :param node_data_2: nx_graph format for the second node to compare

        :return: is the first node equal to the second
        """
        node_1, node_2 = node_data_1.get('node'), node_data_2.get('node')

        operations_do_match = str(node_1) == str(node_2)
        params_do_match = node_1.content.get('params') == node_2.content.get('params')
        nodes_do_match = operations_do_match and params_do_match
        return nodes_do_match

    graphs = (graph_1, graph_2)
    nx_graphs = []
    for graph in graphs:
        nx_graph, nodes = graph_structure_as_nx_graph(graph)
        set_node_attributes(nx_graph, nodes, name='node')
        nx_graphs.append(nx_graph)

    distance = graph_edit_distance(*nx_graphs, node_match=node_match)
    return int(distance)



from os.path import join
from typing import Optional, List, Type
import multiprocessing

from golem.core.log import default_log
from golem.core.dag.graph import Graph, GraphNode
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_node_factory import OptNodeFactory
from golem.core.optimisers.timer import OptimisationTimer
from golem.core.paths import default_data_dir
from golem.structural_analysis.graph_sa.node_sa_approaches import NodeAnalyzeApproach, NodeAnalysis
from golem.structural_analysis.graph_sa.results.sa_analysis_results import SAAnalysisResults
from golem.structural_analysis.graph_sa.sa_requirements import StructuralAnalysisRequirements


class NodesAnalysis:
    """
    This class is for nodes structural analysis within a Graph .
    It takes nodes and approaches to be applied to chosen nodes.
    To define which nodes to analyze pass them to nodes_to_analyze filed
    or all nodes will be analyzed.

    :param objective: objective functions for computing metric values
    :param node_factory: node factory to advise changes from available operations and models
    :param approaches: methods applied to nodes to modify the graph or analyze certain operations.\
    Default: [NodeDeletionAnalyze, NodeReplaceOperationAnalyze]
    :param path_to_save: path to save results to. Default: ~home/Fedot/structural
    """

    def __init__(self, objective: Objective,
                 node_factory: OptNodeFactory,
                 approaches: Optional[List[Type[NodeAnalyzeApproach]]] = None,
                 requirements: Optional[StructuralAnalysisRequirements] = None,
                 path_to_save: Optional[str] = None):

        self.objective = objective
        self.node_factory = node_factory
        self.approaches = approaches
        self.requirements = \
            StructuralAnalysisRequirements() if requirements is None else requirements
        self.log = default_log(self)
        self.path_to_save = \
            join(default_data_dir(), 'structural', 'nodes_structural') if path_to_save is None else path_to_save

    def analyze(self, graph: Graph, results: SAAnalysisResults = None,
                nodes_to_analyze: Optional[List[GraphNode]] = None,
                n_jobs: int = 1, timer: Optional[OptimisationTimer] = None) -> SAAnalysisResults:
        """
        Main method to run the analyze process for every node.

        :param graph: graph object to analyze
        :param results: SA results
        :param nodes_to_analyze: nodes to analyze. Default: all nodes
        :param n_jobs: n_jobs
        :param timer: timer indicating how much time is left for optimization
        :return nodes_results: dict with analysis result per GraphNode
        """

        if not results:
            results = SAAnalysisResults()

        if n_jobs == -1:
            n_jobs = multiprocessing.cpu_count()

        if not nodes_to_analyze:
            self.log.message('Nodes to analyze are not defined. All nodes will be analyzed.')
            nodes_to_analyze = graph.nodes

        node_analysis = NodeAnalysis(approaches=self.approaches,
                                     approaches_requirements=self.requirements,
                                     node_factory=self.node_factory,
                                     path_to_save=self.path_to_save)

        with multiprocessing.Pool(processes=n_jobs) as pool:
            cur_nodes_results = pool.starmap(node_analysis.analyze,
                                       [[graph, node, self.objective, timer]
                                        for node in nodes_to_analyze])

        results.add_results(cur_nodes_results)

        return results



from __future__ import annotations

import os
from copy import deepcopy
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from uuid import uuid4

from golem.core.dag.graph import Graph
from golem.core.log import default_log
from golem.core.optimisers.fitness.fitness import Fitness, null_fitness
from golem.core.optimisers.graph import OptGraph
from golem.serializers.serializer import default_load, default_save

if TYPE_CHECKING:
    from golem.core.optimisers.opt_history_objects.parent_operator import ParentOperator


INDIVIDUAL_COPY_RESTRICTION_MESSAGE = ('`Individual` instance was copied.\n'
                                       'Normally, you don\'t want to do that to keep uid-individual uniqueness.\n'
                                       'If this happened during the optimization process, this misusage '
                                       'should be fixed.')


@dataclass(frozen=True)
class Individual:
    graph: Graph
    parent_operator: Optional[ParentOperator] = field(default=None)
    metadata: Dict[str, Any] = field(default_factory=dict)
    native_generation: Optional[int] = None
    fitness: Fitness = field(default_factory=null_fitness)
    uid: str = field(default_factory=lambda: str(uuid4()))

    def set_native_generation(self, native_generation):
        if self.native_generation is None:
            super().__setattr__('native_generation', native_generation)

    def _set_fitness_and_graph(self, fitness: Fitness, updated_graph: Optional[OptGraph] = None):
        super().__setattr__('fitness', fitness)
        if updated_graph is not None:
            super().__setattr__('graph', updated_graph)

    def set_evaluation_result(self, eval_result: Union[GraphEvalResult, Fitness],
                              updated_graph: Optional[OptGraph] = None):
        if self.fitness.valid:
            raise ValueError('The individual has valid fitness and can not be evaluated again.')

        if isinstance(eval_result, Fitness):
            self._set_fitness_and_graph(eval_result, updated_graph)
            return

        self._set_fitness_and_graph(eval_result.fitness, eval_result.graph)
        self.metadata.update(eval_result.metadata)

    @property
    def has_native_generation(self) -> bool:
        return self.native_generation is not None

    @property
    def parents(self) -> List[Individual]:
        if not self.parent_operator:
            return []
        return list(self.parent_operator.parent_individuals)

    @property
    def parents_from_prev_generation(self) -> List[Individual]:
        parents_from_prev_generation = []
        next_parents = self.parents
        for _ in range(1_000_000):
            if not next_parents or all(p.has_native_generation for p in next_parents):
                break
            parents = next_parents
            next_parents = []
            for p in parents:
                next_parents += p.parents
        else:  # After the last iteration.
            raise ValueError(f'The individual {self.uid} has invalid inheritance data.')

        parents_from_prev_generation += next_parents
        return parents_from_prev_generation

    @property
    def operators_from_prev_generation(self) -> List[ParentOperator]:
        if not self.parent_operator:
            return []
        parents_from_prev_generation = self.parents_from_prev_generation
        operators = [self.parent_operator]
        next_parents = self.parents
        for _ in range(1_000_000):
            if next_parents == parents_from_prev_generation:
                break
            parents = next_parents
            next_parents = []
            for p in parents:
                next_parents += p.parents
                operators.append(p.parent_operator)
        else:  # After the last iteration.
            raise ValueError(f'The individual {self.uid} has invalid inheritance data.')

        operators.reverse()
        return operators

    def save(self, json_file_path: Union[str, os.PathLike] = None) -> Optional[str]:
        return default_save(obj=self, json_file_path=json_file_path)

    @staticmethod
    def load(json_str_or_file_path: Union[str, os.PathLike] = None) -> Individual:
        return default_load(json_str_or_file_path)

    def __repr__(self):
        return (f'<Individual {self.uid} | fitness: {self.fitness} | native_generation: {self.native_generation} '
                f'| graph: {self.graph}>')

    def __eq__(self, other: Individual):
        return self.uid == other.uid

    def __copy__(self):
        default_log(self).warning(INDIVIDUAL_COPY_RESTRICTION_MESSAGE, raise_if_test=True)
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        default_log(self).warning(INDIVIDUAL_COPY_RESTRICTION_MESSAGE, raise_if_test=True)
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            object.__setattr__(result, k, deepcopy(v, memo))
        return result


@dataclass
class GraphEvalResult:
    uid_of_individual: str
    fitness: Fitness
    graph: OptGraph  # For the case if evaluation needs to assign some values to the graph.
    metadata: Dict[str, Any] = field(default_factory=dict)

    def __bool__(self):
        return self.fitness.valid



from typing import Sequence, List

from golem.core.optimisers.adaptive.mab_agents.contextual_mab_agent import ContextualMultiArmedBanditAgent
from golem.core.optimisers.adaptive.neural_mab import NeuralMAB
from golem.core.optimisers.adaptive.context_agents import ContextAgentTypeEnum
from golem.core.optimisers.adaptive.common_types import ActType


class NeuralContextualMultiArmedBanditAgent(ContextualMultiArmedBanditAgent):
    """ Neural Contextual Multi-Armed bandit.
    Observations can be encoded with the use of Neural Networks, but still there are some restrictions
    to guarantee convergence. """
    def __init__(self,
                 actions: Sequence[ActType],
                 context_agent_type: ContextAgentTypeEnum,
                 available_operations: List[str],
                 n_jobs: int = 1,
                 enable_logging: bool = True,
                 decaying_factor: float = 1.0):
        super().__init__(actions=actions, n_jobs=n_jobs,
                         context_agent_type=context_agent_type, available_operations=available_operations,
                         enable_logging=enable_logging, decaying_factor=decaying_factor)
        self._agent = NeuralMAB(arms=self._indices,
                                n_jobs=n_jobs)



import itertools
from dataclasses import dataclass
from numbers import Real
from typing import Any, Optional, Callable, Sequence, TypeVar, Dict, Tuple, Union

from golem.core.dag.graph import Graph
from golem.core.log import default_log
from golem.core.optimisers.fitness import Fitness, SingleObjFitness, null_fitness, MultiObjFitness

G = TypeVar('G', bound=Graph, covariant=True)
R = TypeVar('R', contravariant=True)
GraphFunction = Callable[[G], R]
ObjectiveFunction = GraphFunction[G, Fitness]


@dataclass
class ObjectiveInfo:
    """Keeps information about used metrics."""
    is_multi_objective: bool = False
    metric_names: Sequence[str] = ()

    def __str__(self):
        return f'{self.__class__.__name__}(multi={self.is_multi_objective}, metrics={self.metric_names})'

    def format_fitness(self, fitness: Union[Fitness, Sequence[float]]) -> str:
        """Returns formatted fitness string.
        Example for 3 metrics: `<roc_auc=0.542 f1=0.72 complexity=0.8>`"""
        values = fitness.values if isinstance(fitness, Fitness) else fitness
        fitness_info_str = [f'{name}={value:.3f}'
                            if value is not None
                            else f'{name}=None'
                            for name, value in zip(self.metric_names, values)]
        return f"<{' '.join(fitness_info_str)}>"


class Objective(ObjectiveInfo, ObjectiveFunction):
    """Represents objective function for computing metric values
    on Graphs and keeps information about metrics used."""

    def __init__(self,
                 quality_metrics: Union[Callable, Dict[Any, Callable]],
                 complexity_metrics: Optional[Dict[Any, Callable]] = None,
                 is_multi_objective: bool = False,
                 ):
        self._log = default_log(self)
        if isinstance(quality_metrics, Callable):
            quality_metrics = {'metric': quality_metrics}
        self.quality_metrics = quality_metrics
        self.complexity_metrics = complexity_metrics or {}
        metric_names = [str(metric_id) for metric_id, _ in self.metrics]
        ObjectiveInfo.__init__(self, is_multi_objective, metric_names)

    def __call__(self, graph: Graph, **metrics_kwargs: Any) -> Fitness:
        evaluated_metrics = []
        for metric_id, metric_func in self.metrics:
            try:
                metric_value = metric_func(graph, **metrics_kwargs)
                evaluated_metrics.append(metric_value)
            except Exception as ex:
                self._log.error(f'Objective evaluation error for graph {graph} on metric {metric_id}: {ex}')
                return null_fitness()  # fail right away
        return to_fitness(evaluated_metrics, self.is_multi_objective)

    @property
    def metrics(self) -> Sequence[Tuple[Any, Callable]]:
        return list(itertools.chain(self.quality_metrics.items(), self.complexity_metrics.items()))

    def get_info(self) -> ObjectiveInfo:
        return ObjectiveInfo(self.is_multi_objective, self.metric_names)


def to_fitness(metric_values: Optional[Sequence[Real]], multi_objective: bool = False) -> Fitness:
    if metric_values is None:
        return null_fitness()
    elif multi_objective:
        return MultiObjFitness(values=metric_values, weights=1.)
    else:
        return SingleObjFitness(*metric_values)


def get_metric_position(metrics, metric_type):
    for num, metric in enumerate(metrics):
        if isinstance(metric, metric_type):
            return num



import random
from functools import partial
from typing import Union, Sequence, Optional, List, Callable

import numpy as np
from mabwiser.mab import MAB, LearningPolicy, NeighborhoodPolicy
from scipy.special import softmax

from golem.core.dag.graph import Graph
from golem.core.dag.graph_node import GraphNode
from golem.core.optimisers.adaptive.context_agents import ContextAgentsRepository, ContextAgentTypeEnum
from golem.core.optimisers.adaptive.mab_agents.mab_agent import MultiArmedBanditAgent
from golem.core.optimisers.adaptive.operator_agent import ActType, ObsType, ExperienceBuffer


class ContextualMultiArmedBanditAgent(MultiArmedBanditAgent):
    """ Contextual Multi-Armed bandit. Observations can be encoded with simple context agent without
    using NN to guarantee convergence.

    :param actions: types of mutations
    :param context_agent_type: function to convert observation to its embedding. Can be specified as
    ContextAgentTypeEnum or as Callable function.
    :param available_operations: available operations
    :param n_jobs: n_jobs
    :param enable_logging: bool logging flag
    """

    def __init__(self, actions: Sequence[ActType],
                 context_agent_type: Union[ContextAgentTypeEnum, Callable],
                 available_operations: List[str],
                 n_jobs: int = 1,
                 enable_logging: bool = True,
                 decaying_factor: float = 1.0):
        super().__init__(actions=actions, n_jobs=n_jobs, enable_logging=enable_logging,
                         decaying_factor=decaying_factor, is_initial_fit=False)
        self._agent = MAB(arms=self._indices,
                          learning_policy=LearningPolicy.UCB1(alpha=1.25),
                          neighborhood_policy=NeighborhoodPolicy.Clusters(),
                          n_jobs=n_jobs)
        self._context_agent = context_agent_type if isinstance(context_agent_type, Callable) else \
            partial(ContextAgentsRepository.agent_class_by_id(context_agent_type),
                    available_operations=available_operations)
        self._is_fitted = False

    def _initial_fit(self, obs: ObsType):
        """ Initial fit for Contextual Multi-Armed Bandit.
        At this step, all hands are assigned the same weights with the very first context
        that is fed to the bandit. """
        # initial fit for mab
        n = len(self._indices)
        uniform_rewards = [1. / n] * n
        contexts = self.get_context(obs=obs)
        self._agent.fit(decisions=self._indices, rewards=uniform_rewards, contexts=np.tile(contexts, (n, 1)))
        self._is_fitted = True

    def choose_action(self, obs: ObsType) -> ActType:
        if not self._is_fitted:
            self._initial_fit(obs=obs)
        contexts = self.get_context(obs=obs)
        arm = self._agent.predict(contexts=contexts.reshape(1, -1))
        action = self.actions[arm]
        return action

    def get_action_values(self, obs: Optional[ObsType] = None) -> Sequence[float]:
        if not self._is_fitted:
            self._initial_fit(obs=obs)
        contexts = self.get_context(obs)
        prob_dict = self._agent.predict_expectations(contexts=contexts.reshape(1, -1))
        prob_list = [prob_dict[i] for i in range(len(prob_dict))]
        return prob_list

    def get_action_probs(self, obs: Optional[ObsType] = None) -> Sequence[float]:
        return softmax(self.get_action_values(obs=obs))

    def choose_nodes(self, graph: Graph, num_nodes: int = 1) -> Union[GraphNode, Sequence[GraphNode]]:
        subject_nodes = random.sample(graph.nodes, k=num_nodes)
        return subject_nodes[0] if num_nodes == 1 else subject_nodes

    def partial_fit(self, experience: ExperienceBuffer):
        """Continues learning of underlying agent with new experience."""
        obs, arms, processed_rewards = self._get_experience(experience)
        contexts = self.get_context(obs=obs)
        self._agent.partial_fit(decisions=arms, rewards=processed_rewards, contexts=contexts)

    def _get_experience(self, experience: ExperienceBuffer):
        """ Get experience from ExperienceBuffer, process rewards and log. """
        obs, actions, rewards = experience.retrieve_experience()
        arms = [self._arm_by_action[action] for action in actions]
        # there is no need to process rewards as in MAB, since this processing unifies rewards for all contexts
        self._dbg_log(obs, actions, rewards)
        return obs, arms, rewards

    def get_context(self, obs: Union[List[ObsType], ObsType]) -> np.array:
        """ Returns contexts based on specified context agent. """
        if not isinstance(obs, list):
            obs = [obs]
        contexts = []
        for ob in obs:
            if isinstance(ob, list) or isinstance(ob, np.ndarray):
                # to unify type to list
                contexts.append(np.array(ob).flatten())
            else:
                context = np.array(self._context_agent(ob))
                # some external context agents can wrap context in an additional array
                contexts.append(context.flatten())
        return np.array(contexts)



import os
from pathlib import Path
from typing import Optional, Iterable

from golem.core.log import default_log
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory


class HistoryReader:
    """Simplifies reading a bunch of histories from single directory."""

    def __init__(self, save_path: Optional[Path] = None):
        self.log = default_log(self)
        self.save_path = save_path or Path("results")
        self.save_path.mkdir(parents=True, exist_ok=True)

    def load_histories(self) -> Iterable[OptHistory]:
        """Iteratively loads saved histories one-by-ony."""
        num_histories = 0
        total_individuals = 0
        for history_path in HistoryReader.traverse_histories(self.save_path):
            history = OptHistory.load(history_path)
            num_histories += 1
            total_individuals += sum(map(len, history.generations))
            yield history

        if num_histories == 0 or total_individuals == 0:
            raise ValueError(f'Could not load any individuals.'
                             f'Possibly, path {self.save_path} does not exist or is empty.')
        else:
            self.log.info(f'Loaded {num_histories} histories '
                          f'with {total_individuals} individuals in total.')

    @staticmethod
    def traverse_histories(path) -> Iterable[Path]:
        if path.exists():
            # recursive traversal of the save directory
            for root, dirs, files in os.walk(path):
                for history_filename in files:
                    if history_filename.startswith('history'):
                        full_path = Path(root) / history_filename
                        yield full_path



from typing import Dict, Callable, List, Union

OperationParametersMapping = Dict[str, Dict[str, Dict[str, Union[Callable, List, str]]]]


class SearchSpace:
    """
    Args:
        search_space: dictionary with parameters and their search_space
            {'operation_name': {'param_name': {'hyperopt-dist': hyperopt distribution function,
                                               'sampling-scope': [sampling scope],
                                               'type': 'discrete', 'continuous' or 'categorical'}, ...}, ...},

            e.g. ``{'operation_name': {'param1': {'hyperopt-dist': hp.uniformint,
                                                  'sampling-scope': [2, 21]),
                                                  'type': 'discrete'},
                                       'param2': {'hyperopt-dist': hp.loguniform,
                                                  'sampling-scope': [0.001, 1]),
                                                  'type': 'continuous'},
                                       'param3': {'hyperopt-dist': hp.choice,
                                                  'sampling-scope': [['svd', 'lsqr', 'eigen']),
                                                  'type': 'categorical'}...}, ..}
    """

    def __init__(self, search_space: OperationParametersMapping):
        self.parameters_per_operation = search_space

    def get_parameters_for_operation(self, operation_name: str) -> List[str]:
        parameters_list = list(self.parameters_per_operation.get(operation_name, {}).keys())
        return parameters_list


def get_node_operation_parameter_label(node_id: int, operation_name: str, parameter_name: str) -> str:
    # Name with operation and parameter
    op_parameter_name = ''.join((operation_name, ' | ', parameter_name))

    # Name with node id || operation | parameter
    node_op_parameter_name = ''.join((str(node_id), ' || ', op_parameter_name))
    return node_op_parameter_name


def convert_parameters(parameters):
    """
    Function removes labels from dictionary with operations

    Args:
        parameters: labeled parameters

    Returns:
        new_parameters: dictionary without labels of node_id and operation_name
    """

    new_parameters = {}
    for operation_parameter, value in parameters.items():
        # Remove right part of the parameter name
        parameter_name = operation_parameter.split(' | ')[-1]

        if value is not None:
            new_parameters.update({parameter_name: value})

    return new_parameters



from abc import abstractmethod
from typing import Sequence, Any, Optional, Tuple

import numpy as np

from golem.utilities.data_structures import Comparable


class Fitness(Comparable):
    """Abstracts comparable fitness values that can be in invalid state.

    Importantly, Fitness comparison is semantic: `more-than` means `better-than`.
    Fitness can be compared using standard operators ``>``, ``<``, ``>=``, etc.

    Fitness comparison handles invalid fitness: invalid fitness is never better
    than any other fitness. Fitness implementations must ensure this contract.

    Default Fitness comparison is lexicographic (even for multi-objective fitness,
    to ensure total ordering). For proper comparison of multi-objective fitness
    use method `dominates`.
    """

    @property
    def value(self) -> Optional[float]:
        """Return primary fitness value"""
        return self.values[0]

    @property
    @abstractmethod
    def values(self) -> Sequence[float]:
        """Return individual metric values.
        Returned values are already weighted, if weights are used."""
        raise NotImplementedError()

    @values.setter
    @abstractmethod
    def values(self, new_values: Optional[Sequence[float]]):
        """Assign individual metric values. Accepts unweighted values."""
        raise NotImplementedError()

    @values.deleter
    @abstractmethod
    def values(self):
        """Clear internal metric values."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def weights(self) -> Sequence[float]:
        """Return weights used for weighting individual metrics."""
        raise NotImplementedError()

    @property
    @abstractmethod
    def valid(self) -> bool:
        """Assess if a fitness is valid or not."""
        raise NotImplementedError()

    def dominates(self, other: 'Fitness', selector: Any = None) -> bool:
        """Implementation-specific test for fitness domination.
        By the default behaves same as less-than operator for valid fitness.
        Less means worse; so the better fitness is a dominating one.

        :param other: another fitness for dominates test.
        :param selector: optionally specifies which objectives of the fitness to use.
        """
        return self > other

    def reset(self):
        del self.values

    def __lt__(self, other: 'Fitness') -> bool:
        """'Less-than' for fitness means 'worse-than'.
        NB: in the case of both invalid the other takes precedence
        """
        if not self.valid:
            # invalid self is worse
            return True
        elif not other.valid:
            # valid self is NOT worse than invalid other
            return False
        # if both are valid then compare normally
        return is_metric_worse(self.values, other.values)  # lexicographic comparison

    def __hash__(self) -> int:
        # try to avoid numeric precision errors in hash comparisons
        vals = tuple(np.round(value, 8) if value is not None else None
                     for value in self.values)
        return hash(vals)

    def __str__(self):
        """Return the values of the Fitness object."""
        return str(self.values if self.valid else tuple())

    def __repr__(self):
        """Return the Python code to build a copy of the object."""
        return "%s.%s%r" % (self.__module__, self.__class__.__name__,
                            tuple(self.values) if self.valid else tuple())

    def __eq__(self, other: 'Fitness') -> bool:
        return (isinstance(other, self.__class__) and
                self.valid and other.valid and
                self.allclose(self.values, other.values))

    def __bool__(self) -> bool:
        return self.valid

    @staticmethod
    def allclose(values1, values2) -> bool:
        return np.allclose(values1, values2, rtol=1e-8, atol=1e-10)


class SingleObjFitness(Fitness):
    """Single-objective fitness with optional supplementary values
    for distinguishing cases when primary fitness values are equal.
    This fitness implements lexicographic comparison on its fitness values.

    :param primary_value: Primary fitness metric, may be None. It determines if fitness is valid.
    :param supplementary_values: Define supplementary metrics, must not be None.
    """

    def __init__(self, primary_value: Optional[float] = None, *supplementary_values: float):
        self._values: Tuple = (primary_value, *supplementary_values)

    @property
    def values(self) -> Sequence[float]:
        return self._values

    @values.setter
    def values(self, new_values: Optional[Sequence[float]]):
        if new_values is None:
            self.reset()
        if any(secondary_value is None for secondary_value in new_values[1:]):
            raise ValueError('Secondary values must not be None for prioritized fitness')
        self._values = tuple(new_values)

    @values.deleter
    def values(self):
        self._values = (None,)

    @property
    def weights(self) -> Sequence[float]:
        # Return default weights
        return (1.,) * len(self.values)

    @property
    def valid(self) -> bool:
        return self._values[0] is not None

    def __hash__(self) -> int:
        # __hash__ required explicit super() call
        return super().__hash__()

    def __str__(self) -> str:
        # For single objective return only the primary value
        return str(round(self.value, 4)) if self.value is not None else 'null_fitness'


def null_fitness() -> SingleObjFitness:
    """Alias for creating default-initialised single-value fitness."""
    return SingleObjFitness(primary_value=None)


def is_metric_worse(left_value, right_value) -> bool:
    if isinstance(left_value, Fitness):
        # Fitness object already handles metric comparison in the right way
        return left_value < right_value
    else:
        # Less is better -- minimisation task on raw metric values
        return left_value > right_value



from abc import ABC
from typing import TypeVar, Generic

from golem.core.dag.graph import Graph
from golem.core.optimisers.fitness import Fitness
from .objective import Objective

G = TypeVar('G', bound=Graph, covariant=True)


class ObjectiveEvaluate(ABC, Generic[G]):
    """Defines how Objective must be evaluated on Graphs.

     Responsibilities:
     - Graph-specific evaluation policy: typically, Graphs require some kind of evaluation
     before Objective could be estimated on them. E.g. Machine-learning pipelines must be
     fit on train data before they could be evaluated on the test data.
     - Objective-specific estimation: typically objectives require additional parameters
     besides Graphs for estimation, e.g. test data for estimation of prediction quality.
     - Optionally, compute additional statistics for Graphs (intermediate metrics).

     Default implementation is just a closure that calls :param objective: with
      redirected keyword arguments :param objective_kwargs:
    """

    def __init__(self, objective: Objective, eval_n_jobs: int = 1, **objective_kwargs):
        self._objective = objective
        self._objective_kwargs = objective_kwargs
        self._eval_n_jobs = eval_n_jobs

    def __call__(self, graph: G) -> Fitness:
        """Provides functional interface for ObjectiveEvaluate."""
        return self.evaluate(graph)

    @property
    def eval_n_jobs(self) -> int:
        return self._eval_n_jobs

    @eval_n_jobs.setter
    def eval_n_jobs(self, n_jobs: int):
        self._eval_n_jobs = n_jobs

    def evaluate(self, graph: G) -> Fitness:
        """Evaluate graph and compute its fitness."""
        return self._objective(graph, **self._objective_kwargs)

    def evaluate_intermediate_metrics(self, graph: G):
        """Compute intermediate metrics for each graph node and store it there."""
        pass



from copy import deepcopy
from random import choice
from typing import Sequence, Union, Any

from golem.core.constants import MAX_GRAPH_GEN_ATTEMPTS
from golem.core.dag.graph import Graph
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import Crossover
from golem.core.optimisers.genetic.operators.elitism import Elitism
from golem.core.optimisers.genetic.operators.inheritance import Inheritance
from golem.core.optimisers.genetic.operators.mutation import Mutation
from golem.core.optimisers.genetic.operators.operator import PopulationT, EvaluationOperator
from golem.core.optimisers.genetic.operators.regularization import Regularization
from golem.core.optimisers.genetic.operators.reproduction import ReproductionController
from golem.core.optimisers.genetic.operators.selection import Selection
from golem.core.optimisers.genetic.parameters.graph_depth import AdaptiveGraphDepth
from golem.core.optimisers.genetic.parameters.operators_prob import init_adaptive_operators_prob
from golem.core.optimisers.genetic.parameters.population_size import init_adaptive_pop_size, PopulationSize
from golem.core.optimisers.objective.objective import Objective
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.core.optimisers.populational_optimizer import PopulationalOptimizer


class EvoGraphOptimizer(PopulationalOptimizer):
    """
    Multi-objective evolutionary graph optimizer named GPComp
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Sequence[Union[Graph, Any]],
                 requirements: GraphRequirements,
                 graph_generation_params: GraphGenerationParams,
                 graph_optimizer_params: GPAlgorithmParameters):
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        # Define genetic operators
        self.regularization = Regularization(graph_optimizer_params, graph_generation_params)
        self.selection = Selection(graph_optimizer_params)
        self.crossover = Crossover(graph_optimizer_params, requirements, graph_generation_params)
        self.mutation = Mutation(graph_optimizer_params, requirements, graph_generation_params)
        self.inheritance = Inheritance(graph_optimizer_params, self.selection)
        self.elitism = Elitism(graph_optimizer_params)
        self.operators = [self.regularization, self.selection, self.crossover,
                          self.mutation, self.inheritance, self.elitism]
        self.reproducer = ReproductionController(graph_optimizer_params, self.selection, self.mutation, self.crossover)

        # Define adaptive parameters
        self._pop_size: PopulationSize = init_adaptive_pop_size(graph_optimizer_params, self.generations)
        self._operators_prob = init_adaptive_operators_prob(graph_optimizer_params)
        self._graph_depth = AdaptiveGraphDepth(self.generations,
                                               start_depth=requirements.start_depth,
                                               max_depth=requirements.max_depth,
                                               max_stagnation_gens=graph_optimizer_params.adaptive_depth_max_stagnation,
                                               adaptive=graph_optimizer_params.adaptive_depth)

        # Define initial parameters
        self.requirements.max_depth = self._graph_depth.initial
        self.graph_optimizer_params.pop_size = self._pop_size.initial
        self.initial_individuals = [Individual(graph, metadata=requirements.static_individual_metadata)
                                    for graph in self.initial_graphs]

    def _initial_population(self, evaluator: EvaluationOperator):
        """ Initializes the initial population """
        # Adding of initial assumptions to history as zero generation
        self._update_population(evaluator(self.initial_individuals), 'initial_assumptions')
        pop_size = self.graph_optimizer_params.pop_size

        if len(self.initial_individuals) < pop_size:
            self.initial_individuals = self._extend_population(self.initial_individuals, pop_size)
            # Adding of extended population to history
            self._update_population(evaluator(self.initial_individuals), 'extended_initial_assumptions')

    def _extend_population(self, pop: PopulationT, target_pop_size: int) -> PopulationT:
        verifier = self.graph_generation_params.verifier
        extended_pop = list(pop)
        pop_graphs = [ind.graph for ind in extended_pop]

        # Set mutation probabilities to 1.0
        initial_req = deepcopy(self.requirements)
        initial_req.mutation_prob = 1.0
        self.mutation.update_requirements(requirements=initial_req)

        for iter_num in range(MAX_GRAPH_GEN_ATTEMPTS):
            if len(extended_pop) == target_pop_size:
                break
            new_ind = self.mutation(choice(pop))
            if new_ind:
                new_graph = new_ind.graph
                if new_graph not in pop_graphs and verifier(new_graph):
                    extended_pop.append(new_ind)
                    pop_graphs.append(new_graph)
        else:
            self.log.warning(f'Exceeded max number of attempts for extending initial graphs, stopping.'
                             f'Current size {len(pop)}, required {target_pop_size} graphs.')

        # Reset mutation probabilities to default
        self.mutation.update_requirements(requirements=self.requirements)
        return extended_pop

    def _evolve_population(self, evaluator: EvaluationOperator) -> PopulationT:
        """ Method realizing full evolution cycle """

        # Defines adaptive changes to algorithm parameters
        #  like pop_size and operator probabilities
        self._update_requirements()

        # Regularize previous population
        individuals_to_select = self.regularization(self.population, evaluator)
        # Reproduce from previous pop to get next population
        new_population = self.reproducer.reproduce(individuals_to_select, evaluator)

        # Adaptive agent experience collection & learning
        # Must be called after reproduction (that collects the new experience)
        experience = self.mutation.agent_experience
        experience.collect_results(new_population)
        self.mutation.agent.partial_fit(experience)

        # Use some part of previous pop in the next pop
        new_population = self.inheritance(self.population, new_population)
        new_population = self.elitism(self.generations.best_individuals, new_population)

        return new_population

    def _update_requirements(self):
        if not self.generations.is_any_improved:
            self.graph_optimizer_params.mutation_prob, self.graph_optimizer_params.crossover_prob = \
                self._operators_prob.next(self.population)
            self.log.info(
                f'Next mutation proba: {self.graph_optimizer_params.mutation_prob}; '
                f'Next crossover proba: {self.graph_optimizer_params.crossover_prob}')
        self.graph_optimizer_params.pop_size = self._pop_size.next(self.population)
        self.requirements.max_depth = self._graph_depth.next()
        self.log.info(
            f'Next population size: {self.graph_optimizer_params.pop_size}; '
            f'max graph depth: {self.requirements.max_depth}')

        # update requirements in operators
        for operator in self.operators:
            operator.update_requirements(self.graph_optimizer_params, self.requirements)



from copy import deepcopy
from typing import Optional, Dict, Any, Iterable

import networkx as nx

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.dag.graph_node import GraphNode
from golem.core.optimisers.graph import OptGraph, OptNode


class BaseNetworkxAdapter(BaseOptimizationAdapter[nx.DiGraph]):
    """Base class for adaptation of networkx.DiGraph to optimization graph.
    Allows to use NetworkX directed graphs with our optimizers.

    For custom networkx adapters overwrite methods responsible for
    transformation of single nodes (`_node_adapt` & `_node_restore`).
    """

    def __init__(self):
        super().__init__(base_graph_class=nx.DiGraph)

    def _node_restore(self, node: GraphNode) -> Dict:
        """Transforms GraphNode to dict of NetworkX node attributes.
        Override for custom behavior."""
        parameters = {}
        if hasattr(node, 'parameters'):
            parameters = deepcopy(node.parameters)

        if node.name:
            parameters['name'] = node.name

        return parameters

    def _node_adapt(self, data: Dict) -> OptNode:
        """Transforms a dict of NetworkX node attributes to GraphNode.
        Override for custom behavior."""
        data = deepcopy(data)
        name = data.pop('name', None)
        return OptNode(content={'name': name, 'params': data})

    def _adapt(self, adaptee: nx.DiGraph) -> OptGraph:
        mapped_nodes = {}

        def map_predecessors(node_id) -> Iterable[OptNode]:
            for pred_id in adaptee.predecessors(node_id):
                yield mapped_nodes[pred_id]

        # map nodes
        for node_id, node_data in adaptee.nodes.items():
            # transform node
            node = self._node_adapt(node_data)
            mapped_nodes[node_id] = node

        # map parent nodes
        for node_id, node in mapped_nodes.items():
            # append its parent edges
            node.nodes_from = map_predecessors(node_id)

        return OptGraph(mapped_nodes.values())

    def _restore(self, opt_graph: OptGraph, metadata: Optional[Dict[str, Any]] = None) -> nx.DiGraph:
        nx_graph = nx.DiGraph()
        nx_node_data = {}

        # add nodes
        for node in opt_graph.nodes:
            nx_node_data[node.uid] = self._node_restore(node)
            nx_graph.add_node(node.uid)

        # add edges
        for node in opt_graph.nodes:
            for parent in node.nodes_from:
                nx_graph.add_edge(parent.uid, node.uid)

        # add nodes ad labels
        nx.set_node_attributes(nx_graph, nx_node_data)

        return nx_graph


_NX_NODE_KEY = 'data'


class DumbNetworkxAdapter(BaseNetworkxAdapter):
    """Simple version of networkx adapter that just stores
    `OptNode` as an attribute of NetworkX graph node."""

    def _node_restore(self, node: GraphNode) -> Dict:
        return {_NX_NODE_KEY: node}

    def _node_adapt(self, data: Dict) -> OptNode:
        return data[_NX_NODE_KEY]


class BanditNetworkxAdapter(BaseNetworkxAdapter):
    """ Classic networkx adapter with nodes indexes in names instead of uids.
    It is needed since some frameworks (e.g. karateclub) have asserts in which node
    names should consist only of its indexes.
    """
    def _restore(self, opt_graph: OptGraph, metadata: Optional[Dict[str, Any]] = None) -> nx.DiGraph:
        nx_graph = nx.DiGraph()
        nx_node_data = {}

        # add nodes
        for node in opt_graph.nodes:
            nx_node_data[node.uid] = self._node_restore(node)
            nx_graph.add_node(opt_graph.nodes.index(node))

        # add edges
        for node in opt_graph.nodes:
            for parent in node.nodes_from:
                nx_graph.add_edge(opt_graph.nodes.index(parent), opt_graph.nodes.index(node))

        # add nodes ad labels
        nx.set_node_attributes(nx_graph, nx_node_data)

        return nx_graph
